# pylint: disable=unused-import

import re
import pandas as pd
from typing import List, Tuple

from lexnlp.extract.common.annotations.court_annotation import CourtAnnotation
from lexnlp.utils.lines_processing.line_processor import LineProcessor, LineSplitParams, LineOrPhrase
from lexnlp.utils.lines_processing.phrase_finder import PhraseFinder, PhraseMatch

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class ParserInitParams:
    """
    UniversalCourtsParser initialization parameters
    """
    def __init__(self):
        self.court_pattern_checker = None  # type: re
        self.column_names = {
            'type': 'Court Type',
            'name': 'Court Name',
            'jurisdiction': 'Jurisdiction',
            'alias': 'Alias'
        }
        self.dataframe_paths = []  # type: List[str]
        self.split_ptrs = None   # type: LineSplitParams
        self.key_word_preproc_func = None  # def (text:str) -> str


class MatchFound:
    def __init__(self, subset, entry_start: int, entry_end: int, text: str):
        self.subset = subset
        self.is_exact = len(subset) == 1
        self.court_name = None
        self.jurisdiction = None
        self.court_type = None
        self.entry_start = entry_start
        self.entry_end = entry_end
        self.text = text

    def __repr__(self):
        length = 'nil' if self.subset is None else str(len(self.subset))
        court_name = 'court name: ' + self.court_name if self.court_name is not None else ''
        court_type = 'court type: ' + self.court_type if self.court_type is not None else ''
        court_jur = 'jurisdiction: ' + self.jurisdiction \
            if self.jurisdiction is not None else ''
        return ' '.join(['Exact: ', self.is_exact, '[' + length + ']',
                         court_name, court_type, court_jur])

    def make_sort_key(self):
        return 0 if self.is_exact else 10 if len(self.subset) == 0 else len(self.subset)


class UniversalCourtsParser:
    """
    The class describes a "constructor" for building locale (and region) specific
    parsers, that find reference to courts within the text.

    Use the parse() method to find all reference to courts from the
    text provided.
    Each reference is a dictionary with two keys:
    - "attrs" key leads to the "coordinates" (starting and ending characters) of the
      occurrence within the provided text
    - "tags" key leads to another dictionary, which contains:
      - court official name
      - court's jurisdiction ...

    In order to parse the text you are supposed to create your locale (or region) specific instance of
    UniversalCourtsParser. See the constructor below:
    """

    def __init__(self, ptrs: ParserInitParams):
        """
        :param ptrs.court_pattern_checker: a regex or None, the parser skips the phrase if pattern doesn't match the phrase
        :param ptrs.column_names['type']: "Court Type", e.g. 'Federal District Court'
        :param ptrs.column_names['name']: "Court Name", e.g. 'Southern Georgia District Court'
        :param ptrs.column_names['jurisdiction']: "Jurisdiction", e.g. 'Federal'
        :param ptrs.column_names['alias']: "Alias", e.g. 'C.D. Cal'
        :param ptrs.dataframe_paths: like ['data/us_courts.csv', ...]
        :param ptrs.split_ptrs: phrase splitting processor parameters, see LineProcessor class
        :param ptrs.key_word_preproc_func: a function used to pre-process column values used in text search

        dataframe_paths is a collection of *.CSV files that contain the data like:

        | Jurisdiction || Court Type         || Court Name               || ... |
        | Federal      || Verfassungsgericht || Bundesverfassungsgericht || ... |

        The column 'Court Name' (you may provide another column name instead of Court Name
        in param: court_name_column) should contain unique values that precisely identify each
        of the court given.

        The columns 'Court Type' (param: court_type_column) and 'Jurisdiction'
        (param: jurisdiction_column) in couple may or may not precisely identify the court given.

        At least this parser can identify the court's type and return the annotation that
        neither specifies the court's name nor jurisdiction

        The court_pattern_checker parameter speeds up the parsing process:
        - the whole text or the line would be skipped if this line doesn't match the court_pattern_checker
        E.g., you can pass re.compile('court', re.IGNORECASE) for searching courts' annotations
        for the En locale

        The split_ptrs specify how the parser splits the text into phrases.
        Each phrase can contain zero ore one court annotations. See LineProcessor class.
        For a courts parser phrase bounds usually include punctuation (.,;!?) and conjunctions
        (and, or) or (und, oder)

        The example function for key_word_preproc_func is:
        def preproc_func(text):
             return re.sub('e$', '[e]?', text)
        """

        self.phrase_match_pattern = None if ptrs.court_pattern_checker is None \
            else ptrs.court_pattern_checker
        self.court_type_column = ptrs.column_names['type']
        self.court_name_column = ptrs.column_names['name']
        self.court_alias_column = ptrs.column_names['alias']
        self.jurisdiction_column = ptrs.column_names['jurisdiction']
        self.proc = LineProcessor(line_split_params=ptrs.split_ptrs)
        self.annotations = []  # type: List[CourtAnnotation]
        self.courts = None
        self.load_courts(ptrs.dataframe_paths)
        self.locale = None

        # unique columns
        self.finder_court_name = PhraseFinder(UniversalCourtsParser.get_unique_col_values(
            self.courts[self.court_name_column]), ptrs.key_word_preproc_func)
        self.finder_court_alias = None if len(self.court_alias_column) == 0 else \
            PhraseFinder(UniversalCourtsParser.get_unique_col_values(
                self.courts[self.court_alias_column]), ptrs.key_word_preproc_func)

        # non-unique columns
        self.finder_court_type = PhraseFinder(UniversalCourtsParser.get_unique_col_values(
            self.courts[self.court_type_column]), ptrs.key_word_preproc_func)
        self.finder_jur = PhraseFinder(UniversalCourtsParser.get_unique_col_values(
            self.courts[self.jurisdiction_column]), ptrs.key_word_preproc_func)

    def parse(self, text: str, locale: str = None) -> List[CourtAnnotation]:
        """
        :param text: the text being processed
        :param locale: 'En', 'Es', ...
        :return: annotations - List[dict]

        Here is an example of the method's call:
        ret = processor.parse("Bei dir läuft, deine Verfassungsgerichtshof des Freistaates Sachsen rauchen Joints vor der Kamera")

        ret[0]['attrs'] = {'start': 14, 'end': 97}
        ret[0]['tags'] = {'Extracted Entity Type': 'court',
            'Extracted Entity Court Name': 'Verfassungsgerichtshof des Freistaates Sachsen',
            'Extracted Entity Court Type': 'Verfassungsgericht',
            'Extracted Entity Court Jurisdiction': 'Sachsen'}
        """
        self.annotations = []
        self.locale = locale
        self.find_courts_by_alias_in_whole_text(text)

        # if the whole text doesn't contain the key word (gericht) - skip all the following
        if self.phrase_match_pattern is not None:
            if self.phrase_match_pattern.search(text, re.IGNORECASE) is None:
                return self.annotations

        for phrase in self.proc.split_text_on_line_with_endings(text):
            # if the phrase doesn't contain the key word (e.g., gericht for deutsche) - skip the phrase
            if self.phrase_match_pattern is not None:
                if self.phrase_match_pattern.search(phrase.text, re.IGNORECASE) is None:
                    continue
            self.find_court_by_any_key(phrase)

        return self.annotations

    def load_courts(self, dataframe_paths: List[str]):
        frames = []
        dtypes = {self.court_type_column: str,
                  self.court_name_column: str,
                  self.jurisdiction_column: str}
        if self.court_alias_column:
            dtypes[self.court_alias_column] = str

        for path in dataframe_paths:
            frame = pd.read_csv(path, encoding="utf-8", error_bad_lines=False,
                                converters=dtypes)
            frames.append(frame)
        self.courts = pd.concat(frames)

    def find_courts_by_alias_in_whole_text(self, text: str) -> None:
        if self.finder_court_alias is None:
            return
        for m in self.finder_court_alias.find_word(text):
            alias = m[0]
            rows = self.courts.loc[self.courts[self.court_alias_column] == alias]
            match_found = MatchFound(rows, m[1], m[2], text[m[1]:m[2]])
            self.add_annotation(match_found)

    def find_court_by_any_key(self, phrase: LineOrPhrase):
        # find by court names
        matches = []
        matches += self.find_court_by_name(phrase)
        matches += self.find_court_by_type_and_jurisdiction(phrase)
        matches = [m for m in matches if m is not None]
        if len(matches) == 0:
            return
        # find the best match
        matches.sort(key=lambda m: m.make_sort_key())
        self.add_annotation(matches[0])

    def find_court_by_name(self, phrase: LineOrPhrase) -> List[MatchFound]:
        match = self.find_court_by_key_column(phrase, self.finder_court_name,
                                                self.court_name_column)
        if match is None:
            return []

        match[0].court_name = match[1][0][0]
        return [match[0]]

    def find_court_by_key_column(self, phrase: LineOrPhrase,
                                 phrase_finder: PhraseFinder,
                                 column: str) -> Tuple[MatchFound, List[PhraseMatch]]:
        found_substrings = phrase_finder.find_word(phrase.text, True)
        if len(found_substrings) == 0:
            return None
        subset = self.courts.loc[self.courts[column] == found_substrings[0][0]]
        if len(subset) == 0:
            return None

        start = found_substrings[0][1]
        end = found_substrings[0][2]
        match = MatchFound(subset,
                           phrase.start + start,
                           phrase.start + end,
                           phrase.text[start:end])
        return match, found_substrings

    def find_court_by_type_and_jurisdiction(self, phrase: LineOrPhrase) -> List[MatchFound]:
        court_types = self.finder_court_type.find_word(phrase.text, True)
        if len(court_types) == 0:
            return []

        court_jurs = self.finder_jur.find_word(phrase.text, True)
        if len(court_types) != 1 or len(court_jurs) > 1:
            # special case: 2 ore more courts within the same phrase
            # (without commas or conjuctions)
            matches = []
            for ct in court_types:
                m = MatchFound([],
                               phrase.start + ct[1],
                               phrase.start + ct[2],
                               phrase.text[ct[1]:ct[2]])
                m.court_type = ct[0]
                m.court_name = ct[0]
                matches.append(m)
            return matches

        if len(court_jurs) == 0:
            subset = self.courts.loc[self.courts[self.court_type_column] == court_types[0][0]]
        else:
            subset = self.courts.loc[(self.courts[self.court_type_column] == court_types[0][0]) &
                                     (self.courts[self.jurisdiction_column] == court_jurs[0][0])]

        match = MatchFound(subset,
                           phrase.start,
                           phrase.start + court_types[0][2],
                           phrase.text[0:court_types[0][2]])
        if len(subset) != 1:
            match.court_name = court_types[0][0]
            match.court_type = court_types[0][0]
        return [match]

    def add_annotation(self, match: MatchFound):
        mlen = len(match.subset)

        name = match.subset[self.court_name_column].values[0] \
            if match.is_exact else \
            match.court_name if match.court_name is not None else \
            match.subset[self.court_name_column].values[0] if mlen > 0 else ''

        court_type = match.subset[self.court_type_column].values[0] \
            if match.is_exact else \
            match.court_type if match.court_type is not None else \
            match.subset[self.court_type_column].values[0] if mlen > 0 else ''

        jurisdiction = match.subset[self.jurisdiction_column].values[0] \
            if match.is_exact else \
            match.jurisdiction if match.jurisdiction is not None else \
                match.subset[self.jurisdiction_column].values[0] if mlen > 0 else ''

        ant = CourtAnnotation(name=name,
                              coords=(match.entry_start, match.entry_end),
                              locale=self.locale,
                              text=match.text)
        ant.jurisdiction = jurisdiction
        ant.court_type = court_type
        self.annotations.append(ant)

    @staticmethod
    def get_unique_col_values(col_values):
        return [c for c in col_values.unique() if c]



from typing import List, Match, Callable
import regex as re
from lexnlp.extract.common.definitions.universal_definition_parser import UniversalDefinitionsParser
from lexnlp.extract.common.pattern_found import PatternFound

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class CommonDefinitionPatterns:
    reg_semicolon = re.compile("([\"'“„])(?:(?=(\\\\?))\\2.)*?\\1(?=:)", re.UNICODE | re.IGNORECASE)
    reg_quoted = re.compile("([\"'“„])(?:(?=(\\\\?))\\2.)*?\\1", re.UNICODE | re.IGNORECASE)
    reg_acronyms = re.compile(r"\(\p{Lu}\p{L}*\p{Lu}\)", re.UNICODE)

    @staticmethod
    def match_acronyms(phrase: str) -> List[PatternFound]:
        """
        :param phrase: rompió el silencio tras ser despedido del Canal del Fútbol (CDF).
        :return: {name: 'CDF', probability: 100, ...}
        """
        defs = []
        for match in CommonDefinitionPatterns.reg_acronyms.finditer(phrase):
            acr_start = CommonDefinitionPatterns.get_acronym_words_start(phrase, match)
            if acr_start < 0:
                continue
            df = PatternFound()
            df.name = match.group().strip('() ')
            df.start = acr_start
            df.end = match.start() - 1
            df.probability = 100
            defs.append(df)
        return defs

    @staticmethod
    def get_acronym_words_start(phrase: str, match: Match) -> int:
        """
        each acronym match should be preceded by capitalized words that start from the same letters
        :param phrase: "rompió el silencio tras ser despedido del Canal del Fútbol (CDF). "
        :param match: "(CDF)" Match object for this example
        :return: start letter (42 for this case) index or -1
        """
        proc = UniversalDefinitionsParser.basic_line_processor
        name = match.group().strip('() ').upper()
        start = match.start()
        words = proc.split_text_on_words(phrase[:start])
        if len(words) < 2:
            return -1

        mistakes = 0
        uppercases = 0
        acr_index = len(name) - 1
        acr_start = words[-1].start

        for i in range(len(words) - 1, -1, -1):
            if words[i].is_separator:
                continue
            l = words[i].text[0]
            l_upper = l.upper()
            is_upper = l_upper == l
            if is_upper:
                uppercases += 1
            is_correct = name[acr_index] == l_upper
            if not is_correct:
                mistakes += 1
                if mistakes > 1:
                    return -1
                continue
            acr_start = words[i].start
            acr_index -= 1
            if acr_index < 0:
                break
        return acr_start if uppercases > 1 and acr_index < 0 else -1

    @staticmethod
    def match_es_def_by_semicolon(phrase: str) -> List[PatternFound]:
        """
        :param phrase: "Modern anatomy human": a human of modern anatomy.
        :return: {name: 'Modern anatomy human', probability: 100, ...}
        """
        prob = 100
        defs = []

        for match in CommonDefinitionPatterns.reg_semicolon.finditer(phrase):
            df = PatternFound()
            df.name = match.group()
            df.start = 0
            df.end = len(phrase)
            df.probability = prob
            defs.append(df)
            prob = 66

        return defs

    @staticmethod
    def peek_quoted_part(phrase: str,
                         match: Match,
                         start_func: Callable[[str, Match, Match], int],
                         end_func: Callable[[str, Match, Match], int],
                         match_prob: int) -> List[PatternFound]:
        """
        :param phrase: the whole text, may be used for getting the definition's text length
        :param match: the matched part of the phrase that may contain several quote-packed definitions
        :param start_func: (phrase, match, quoted_match) -> definition's start
        :param end_func: (phrase, match, quoted_match) -> definition's end
        :param match_prob: definition's probability
        :return: a list of definitions found or an empty list
        """
        defs = []
        text = match.group()
        quoted_entries = [m for m in CommonDefinitionPatterns.reg_quoted.finditer(text)]
        if len(quoted_entries) == 0:
            return defs
        for entry in quoted_entries:
            df = PatternFound()
            df.name = entry.group()
            df.start = start_func(phrase, match, entry)
            df.end = end_func(phrase, match, entry)
            df.probability = match_prob
            defs.append(df)
        return defs

    @staticmethod
    def collect_regex_matches_with_quoted_chunks(phrase: str, reg: re, prob: int,
                                                 quoted_def_start: Callable[[str, Match, Match], int],
                                                 quoted_def_end: Callable[[str, Match, Match], int],
                                                 def_start: Callable[[str, Match], int],
                                                 def_end: Callable[[str, Match], int]
                                                 ) -> List[PatternFound]:
        """
        First, find all matches by 'reg' ptr
        Second, go through matches
        For each match try to find a set of quoted words
        If found, use them as matches
        Or use the whole match
        :param quoted_def_start: (phrase, match, quoted_match) -> definition's start
        :param quoted_def_end: (phrase, match, quoted_match) -> definition's end
        :param def_start: (phrase, match) -> definition's start
        :param def_end: (phrase, match) -> definition's end
        :return:
        """
        defs = []
        for match in reg.finditer(phrase):
            quoted_matches = \
                CommonDefinitionPatterns.peek_quoted_part(phrase,
                                                          match,
                                                          quoted_def_start,
                                                          quoted_def_end,
                                                          prob)
            if len(quoted_matches) > 0:
                defs += quoted_matches
                continue

            df = PatternFound()
            df.name = match.group()
            df.start = def_start(phrase, match)
            df.end = def_end(phrase, match)
            df.probability = prob
            defs.append(df)

        return defs

    @staticmethod
    def collect_regex_matches(phrase: str, reg: re, prob: int,
                              def_start: Callable[[str, Match], int],
                              def_end: Callable[[str, Match], int]
                              ) -> List[PatternFound]:
        """
        find all matches by 'reg' ptr
        :param quoted_def_start: (phrase, match, quoted_match) -> definition's start
        :param quoted_def_end: (phrase, match, quoted_match) -> definition's end
        :param def_start: (phrase, match) -> definition's start
        :param def_end: (phrase, match) -> definition's end
        :return:
        """
        defs = []
        for match in reg.finditer(phrase):

            df = PatternFound()
            df.name = match.group()
            df.start = def_start(phrase, match)
            df.end = def_end(phrase, match)
            df.probability = prob
            defs.append(df)

        return defs



"""PII extraction for English.

This module implements PII extraction functionality in English.

Todo:
  * http://www.doncio.navy.mil/contentview.aspx?id=2428
"""

# Imports
import regex as re
from typing import Generator

from lexnlp.extract.common.annotations.text_annotation import TextAnnotation
from lexnlp.extract.common.annotations.phone_annotation import PhoneAnnotation
from lexnlp.extract.common.annotations.ssn_annotation import SsnAnnotation

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


SSN_PATTERN = r"""
(?P<block1>[0-9]{3})[\-]?(?P<block2>[0-9]{2})[\-]?(?P<block3>[0-9]{4})
"""
RE_SSN = re.compile(SSN_PATTERN, re.IGNORECASE | re.UNICODE | re.DOTALL | re.VERBOSE)

US_PHONE_PATTERN = r"""
[\(]?(?P<area_code>[0-9]{3})[\)]?
[\-\s]+
(?P<exchange>[0-9]{3})
[\-\s]+
(?P<last4>[0-9]{4})
"""
RE_US_PHONE = re.compile(US_PHONE_PATTERN, re.IGNORECASE | re.UNICODE | re.DOTALL | re.VERBOSE)


def get_ssns(text, return_sources=False) -> Generator:
    """
    Find possible SSN references in the text.
    """

    # Iterate through all potential matches
    for ant in get_ssn_annotations(text):
        if return_sources:
            yield ant.number, ant.text
        else:
            yield ant.number


def get_ssn_annotations(text: str) -> Generator[SsnAnnotation, None, None]:
    for match in RE_SSN.finditer(text):
        # Get individual group matches
        captures = match.capturesdict()
        blocks = [int(captures[f'block{i + 1}'].pop()) for i in range(3)]
        if not all(blocks):
            continue

        ssn = f'{blocks[0]:03d}-{blocks[1]:02d}-{blocks[2]:04d}'
        ant = SsnAnnotation(coords=match.span(),
                            number=ssn,
                            text=match.group())
        yield ant


def get_us_phones(text: str, return_sources=False) -> Generator:
    """
    Find possible telephone numbers in the text.
    """
    for ant in get_us_phone_annotations(text):
        if return_sources:
            yield ant.phone, ant.text
        else:
            yield ant.phone


def get_us_phone_annotations(text: str) \
        -> Generator[PhoneAnnotation, None, None]:
    """
    Find possible telephone numbers in the text.
    """

    # Iterate through all potential matches
    for match in RE_US_PHONE.finditer(text):
        # Get individual group matches
        captures = match.capturesdict()
        phone = "({area_code}) {exchange}-{last4}".format(area_code=captures["area_code"].pop(),
                                                          exchange=captures["exchange"].pop(),
                                                          last4=captures["last4"].pop(),
                                                          )
        ant = PhoneAnnotation(coords=match.span(),
                              phone=phone,
                              text=match.group(0))
        yield ant


def get_pii(text: str, return_sources=False) -> Generator:
    """
    Find possible PII references in the text.
    :param text:
    :param return_sources:
    :return:
    """

    for ssn in get_ssns(text, return_sources):
        if isinstance(ssn, tuple) or isinstance(ssn, list):
            row = ['ssn']
            for v in ssn:
                row.append(v)
            yield tuple(row)
        else:
            yield ('ssn', ssn)

    for phone in get_us_phones(text, return_sources):
        if isinstance(phone, tuple) or isinstance(phone, list):
            row = ['us_phone']
            for v in phone:
                row.append(v)
            yield tuple(row)
        else:
            yield ('us_phone', phone)


def get_pii_annotations(text: str) -> \
        Generator[TextAnnotation, None, None]:
    """
    Find possible PII references in the text.
    """

    yield from get_ssn_annotations(text)
    yield from get_us_phone_annotations(text)



import re
from typing import Generator, List, Union, Tuple

import pandas as pd

from lexnlp.utils.lines_processing.line_processor import LineProcessor

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class DataframeEntityParser(object):
    """
    Class that provides ability to extract entities from a text
    having some collection of entities formed as dataframe.
    By default it means that dataframe has UNIQUE values in those columns you use for search.
    Returns dict of start/end positions of found item in a text
    and other user-defined key-value pairs

    Params:
      - dataframe: pandas.DataFrame with entities collection
      - parse_columns: list or tuple - these columns will be used to search their values in a text
      - result_columns: dict - map like {'dataframe column name to take a value
        corresponding with extracted entity': 'new_column_name'}
      - preformed_entity: dict - initial, static key-value pairs to use for each extracted entity
      - priority_sort_column: str - column name to sort by and get first match
        if multiple results found, otherwise the first matched row will be used
      - priority_sort_ascending: bool - sort order for priority_sort_column
      - cell_values_separator: str or None - multiple values in datafame cell separated by
        that separator
      - unique_column_values: bool - dataframe columns have unique values

    E.g.:
        >>> parse_columns = ('Kurztitel', 'Titel', 'Abkürzung')
        >>> result_columns = {'Titel': 'name'}
        >>> preformed_entity = {'entity_type': 'Laws and Rules',
        >>>                     'source': 'BaFin',
        >>>                     'country': 'Germany'}
        >>> sort_column = 'Titel'
        >>> items = DataframeEntityParser(
        >>>     df, parse_columns, result_columns, preformed_entity, sort_column).parse(text)
    """

    SEARCH_PTN = r'(?:^|\W)({})(?:\W|$)'

    def __init__(self,
                 dataframe,
                 parse_columns,
                 result_columns=None,
                 preformed_entity=None,
                 priority_sort_column=None,
                 priority_sort_ascending=True,
                 cell_values_separator=';',
                 unique_column_values=True,
                 line_processor: LineProcessor = None):
        self.dataframe = dataframe.fillna('')
        self.parse_columns = parse_columns
        self.result_columns = result_columns or {}
        self.preformed_entity = preformed_entity or {}
        self.priority_sort_column = priority_sort_column
        self.priority_sort_ascending = priority_sort_ascending
        self.cell_values_separator = cell_values_separator
        self.unique_column_values = unique_column_values
        self.line_processor = line_processor

        collection_patterns = \
            [(col_name, self.get_collection_ptn(self.dataframe[col_name].values))
            for col_name in parse_columns if col_name]
        self.collection_patterns = {
                c[0]: c[1] for c in collection_patterns if c[1]
            }

    def get_collection_ptn(self, collection):
        """
        Convert list of values to regex pattern
        :param collection: list of entities to search in
        :return: compilled regex pattern
        """
        collection = [c for c in collection if c]
        if not collection:
            return None

        ptn = self.SEARCH_PTN.format(
            '|'.join(re.escape(j) for i in collection for j in i.split(self.cell_values_separator) if i))
        return re.compile(ptn)

    def get_single_result(self, rows):
        """
        By default we mean that all values we filter by in dataframe are UNIQUE, so just take 1st
        Implement your own logic to choose from multiple matched dataframe rows
        """
        if self.priority_sort_column:
            rows = rows.sort_values(by=self.priority_sort_column,
                                    ascending=self.priority_sort_ascending)
        return rows.iloc[0]

    def get_formed_entity(self, match, col_name):
        """
        Get formed entity from matched row in dataframe
        :param match: re.match object
        :param col_name: df column name
        :return: dict
        """
        matched_str = match.groups()[0]
        location_start, location_end = match.span()
        formed_entity = {
            'location_start': location_start,
            'location_end': location_end,
            'source': matched_str
        }
        if self.result_columns:
            matched_rows = self.dataframe[self.dataframe[col_name].str.contains(r'(?:^|;){}(?:$|;)'.format(matched_str), regex=True)]
            if self.unique_column_values:
                matched_row = self.get_single_result(matched_rows)
                for _col_name, new_col_name in self.result_columns.items():
                    formed_entity[new_col_name] = matched_row[_col_name]
            else:
                formed_entity["entities"] = []
                for _, matched_row in matched_rows.iterrows():
                    sub_entity = {}
                    for _col_name, new_col_name in self.result_columns.items():
                        sub_entity[new_col_name] = matched_row[_col_name]
                    formed_entity["entities"].append(sub_entity)

        formed_entity.update(self.preformed_entity)
        return formed_entity

    def get_entities(self, text: str):
        if self.line_processor:
            # split text on sentences and remove linebreaks within sentences
            for sent in self.line_processor.split_text_on_line_with_endings(text):
                for ent in self.get_entities_from_text(sent.text):
                    ent['location_start'] += sent.start
                    ent['location_end'] += sent.start
                    yield ent
        else:
            yield from self.get_entities_from_text(text)

    def get_entities_from_text(self, text: str) -> Generator[dict, None, None]:
        sent_text = text.replace('\n', ' ')
        for col_name, collection_ptn in self.collection_patterns.items():
            for match in collection_ptn.finditer(sent_text):
                yield self.get_formed_entity(match, col_name)

    def get_entity_list(self, text):
        return list(self.get_entities(text))


def get_entities(text: str,
                 config: pd.DataFrame,
                 parse_columns: Union[List[str], Tuple[str]],
                 result_columns: Union[dict, None] = None,
                 preformed_entity: Union[dict, None] = None,
                 priority_sort_column: Union[str, None] = None,
                 priority_sort_ascending: bool = True,
                 cell_values_separator: Union[str, None] = ';',
                 unique_column_values: bool = True) -> Generator:
    """
    Simple wrapper around DataframeEntityParser
    """
    yield from DataframeEntityParser(dataframe=config,
                                     parse_columns=parse_columns,
                                     result_columns=result_columns,
                                     preformed_entity=preformed_entity,
                                     priority_sort_column=priority_sort_column,
                                     priority_sort_ascending=priority_sort_ascending,
                                     cell_values_separator=cell_values_separator,
                                     unique_column_values=unique_column_values).get_entities(text)


def get_entity_list(text: str,
                    config: pd.DataFrame,
                    parse_columns: Union[List[str], Tuple[str]],
                    result_columns: Union[dict, None] = None,
                    preformed_entity: Union[dict, None] = None,
                    priority_sort_column: Union[str, None] = None,
                    priority_sort_ascending: bool = True,
                    cell_values_separator: Union[str, None] = ';',
                    unique_column_values: bool = True) -> List:
    """
    Simple wrapper around DataframeEntityParser
    """
    return DataframeEntityParser(dataframe=config,
                                 parse_columns=parse_columns,
                                 result_columns=result_columns,
                                 preformed_entity=preformed_entity,
                                 priority_sort_column=priority_sort_column,
                                 priority_sort_ascending=priority_sort_ascending,
                                 cell_values_separator=cell_values_separator,
                                 unique_column_values=unique_column_values).get_entity_list(text)



from typing import Callable, Dict, Set, List, Any, Tuple
from enum import Enum

from lexnlp.extract.common.annotation_type import AnnotationType
from lexnlp.extract.en.acts import get_acts_annotations, get_acts
from lexnlp.extract.en.amounts import get_amounts, get_amount_annotations
from lexnlp.extract.en.citations import get_citation_annotations, get_citations
from lexnlp.extract.en.conditions import get_condition_annotations, get_conditions
from lexnlp.extract.en.constraints import get_constraint_annotations, get_constraints
from lexnlp.extract.en.copyright import get_copyright_annotations, get_copyright
from lexnlp.extract.en.courts import get_court_annotations, _get_courts
from lexnlp.extract.en.cusip import get_cusip_annotations, get_cusip
from lexnlp.extract.en.dates import get_date_annotations, get_dates
from lexnlp.extract.en.definitions import get_definition_annotations, get_definitions
from lexnlp.extract.en.distances import get_distance_annotations, get_distances
from lexnlp.extract.en.durations import get_duration_annotations, get_durations
from lexnlp.extract.en.geoentities import get_geoentity_annotations, get_geoentities
from lexnlp.extract.en.money import get_money_annotations, get_money
from lexnlp.extract.en.percents import get_percent_annotations, get_percents
from lexnlp.extract.en.pii import get_pii_annotations, get_pii, get_us_phone_annotations, get_us_phones, get_ssn_annotations, \
    get_ssns
from lexnlp.extract.en.ratios import get_ratio_annotations, get_ratios
from lexnlp.extract.en.regulations import get_regulation_annotations, get_regulations
from lexnlp.extract.en.trademarks import get_trademark_annotations, get_trademarks
from lexnlp.extract.en.urls import get_url_annotations, get_urls

from lexnlp.extract.de.amounts import get_amount_annotations as get_de_amount_annotations
from lexnlp.extract.de.amounts import get_amounts as get_de_amounts
from lexnlp.extract.de.citations import get_citation_annotations as get_de_citation_annotations
from lexnlp.extract.de.citations import get_citations as get_de_citations
from lexnlp.extract.de.copyrights import get_copyright_annotations as get_de_copyright_annotations
from lexnlp.extract.de.copyrights import get_copyrights as get_de_copyrights
from lexnlp.extract.de.court_citations import get_court_citation_annotations as get_de_court_citation_annotations
from lexnlp.extract.de.court_citations import get_court_citations as get_de_court_citations
from lexnlp.extract.de.courts import get_court_annotations as get_de_court_annotations
from lexnlp.extract.de.courts import get_courts as get_de_courts
from lexnlp.extract.de.dates import get_date_annotations as get_de_date_annotations
from lexnlp.extract.de.dates import get_dates as get_de_dates
from lexnlp.extract.de.definitions import get_definition_annotations as get_de_definition_annotations
from lexnlp.extract.de.definitions import get_definitions as get_de_definitions
from lexnlp.extract.de.durations import get_duration_annotations as get_de_duration_annotations
from lexnlp.extract.de.durations import get_durations as get_de_durations
from lexnlp.extract.de.geoentities import get_geoentity_annotations as get_de_geoentity_annotations
from lexnlp.extract.de.geoentities import get_geoentities as get_de_geoentities
from lexnlp.extract.de.laws import get_law_annotations as get_de_law_annotations
from lexnlp.extract.de.laws import get_laws as get_de_laws
from lexnlp.extract.de.percents import get_percent_annotations as get_de_percent_annotations
from lexnlp.extract.de.percents import get_percents as get_de_percents

from lexnlp.extract.es.copyrights import get_copyright_annotations as get_es_copyright_annotations
from lexnlp.extract.es.copyrights import get_copyrights as get_es_copyrights
from lexnlp.extract.es.courts import get_court_annotations as get_es_court_annotations
from lexnlp.extract.es.courts import get_courts as get_es_courts
from lexnlp.extract.es.dates import get_date_annotations as get_es_date_annotations
from lexnlp.extract.es.dates import get_dates as get_es_dates
from lexnlp.extract.es.definitions import get_definition_annotations as get_es_definition_annotations
from lexnlp.extract.es.definitions import get_definitions as get_es_definitions
from lexnlp.extract.es.regulations import get_regulation_annotations as get_es_regulation_annotations
from lexnlp.extract.es.regulations import get_regulations as get_es_regulations

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class ExtractorResultFormat(Enum):
    """
    What output format we expect:
    - fmt_class produces collections of classes, derived from TextAnnotation
    - fmt_dict collects TextAnnotations, then makes dictionaries of them
      ({ 'attrs': { 'start': 105, 'end': 119 }, 'tags': ... })
    - fmt_object gives tuples or other "legacy" formats
    """
    fmt_class = 1
    fmt_dict = 2
    fmt_object = 3


class ExtractingFunction:
    def __init__(self,
                 method: Callable,
                 fact_type: AnnotationType = None,
                 lang: str = '',
                 result_fmt: ExtractorResultFormat = None):
        self.method = method  # <get_acts_annotations>
        self.fact_type = fact_type  # act
        self.language = lang  # "en"
        self.result_fmt = result_fmt  # fmt_class


class FactExtractor:
    """
    Takes text and language.
    Returns collection of entities (acts, amounts, geoentities etc)
    in desired (ExtractorResultFormat) format.

    Some parsers require additional arguments - for example, geolocation lists
    for EN and DE geoentities parsers. Provide these arguments in
    ensure_parser_arguments_en and ensure_parser_arguments_de.
    """

    ALL_ANT_TYPES = set(AnnotationType)
    LANGUAGE_EN = 'en'
    LANGUAGE_DE = 'de'
    LANGUAGE_ES = 'es'
    # { "en": { "class": { "amount": <routine>, ...
    func_by_lang = {}  # type:Dict[str, Dict[ExtractorResultFormat, Dict[AnnotationType, ExtractingFunction]]]

    parser_extra_arguments = {}  # type:Dict[str, Dict[ExtractorResultFormat, Dict[AnnotationType, Tuple]]]

    @staticmethod
    def parse_text(text: str,
                   lang: str,
                   result_fmt: ExtractorResultFormat = ExtractorResultFormat.fmt_class,
                   extract_all: bool = True,
                   include_types: Set[AnnotationType] = None,
                   exclude_types: Set[AnnotationType] = None) -> Dict[AnnotationType, List[Any]]:
        if lang not in FactExtractor.func_by_lang:
            langs = ', '.join([l for l in FactExtractor.func_by_lang])
            raise Exception(f'Language "{lang}" was not found among {langs}')
        lang_extractors = FactExtractor.func_by_lang[lang]
        result_fmt_key = ExtractorResultFormat.fmt_class \
            if result_fmt == ExtractorResultFormat.fmt_dict else result_fmt

        if result_fmt_key not in lang_extractors:
            raise Exception(f'Format "{result_fmt_key.name}" is not supported for {lang}')
        extractors = lang_extractors[result_fmt_key]

        target_types = set()  # type:  Set[AnnotationType]

        if not extract_all and not include_types:
            return {}
        if not extract_all and include_types:
            target_types = include_types
        elif extract_all:
            target_types = FactExtractor.ALL_ANT_TYPES
            if exclude_types:
                target_types -= exclude_types

        if not target_types:
            return {}
        extractors = [extractors.get(t) for t in target_types]
        extractors = [e for e in extractors if e]  # type: List[ExtractingFunction]

        extra_args = FactExtractor.parser_extra_arguments.get(lang)
        if extra_args:
            extra_args = extra_args.get(result_fmt)
        extra_args = extra_args or {}  # type: Dict[AnnotationType, Tuple]

        facts = {}  # type: Dict[AnnotationType, List[Any]]
        for extractor in extractors:
            extras = extra_args.get(extractor.fact_type)
            func_args = (text,) + extras if extras else (text,)
            typed_facts = list(extractor.method(*func_args))
            if not typed_facts:
                continue
            if result_fmt == ExtractorResultFormat.fmt_dict:
                typed_facts = [f.to_dictionary() for f in typed_facts]
            facts[extractor.fact_type] = typed_facts

        return facts

    @staticmethod
    def ensure_parser_arguments_en(
            geo_config: List[Any] = None) -> None:
        for fmt in ExtractorResultFormat:
            FactExtractor.ensure_parser_arguments(FactExtractor.LANGUAGE_EN,
                                                  fmt,
                                                  AnnotationType.geoentity,
                                                  (geo_config,))

    @staticmethod
    def ensure_parser_arguments_de(
            geo_config: List[Any] = None) -> None:
        for fmt in ExtractorResultFormat:
            FactExtractor.ensure_parser_arguments(FactExtractor.LANGUAGE_DE,
                                                  fmt,
                                                  AnnotationType.geoentity,
                                                  (geo_config,))

    @staticmethod
    def initialize():
        FactExtractor.initialize_en()
        FactExtractor.initialize_de()
        FactExtractor.initialize_es()

    @staticmethod
    def ensure_parser_arguments(lang: str,
                                result_fmt: ExtractorResultFormat,
                                ant_type: AnnotationType,
                                func_args: Tuple) -> None:
        lang_args = FactExtractor.parser_extra_arguments.get(lang)
        if not lang_args:
            lang_args = {}
            FactExtractor.parser_extra_arguments[lang] = lang_args

        fmt_args = lang_args.get(result_fmt)
        if not fmt_args:
            fmt_args = {}
            lang_args[result_fmt] = fmt_args

        fmt_args[ant_type] = func_args

    @staticmethod
    def initialize_en():
        all_functs = [
            ExtractingFunction(method=get_acts_annotations, fact_type=AnnotationType.act,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_acts, fact_type=AnnotationType.act,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_amount_annotations, fact_type=AnnotationType.amount,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_amounts, fact_type=AnnotationType.amount,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_citation_annotations, fact_type=AnnotationType.citation,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_citations, fact_type=AnnotationType.citation,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_condition_annotations, fact_type=AnnotationType.condition,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_conditions, fact_type=AnnotationType.condition,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_constraint_annotations, fact_type=AnnotationType.constraint,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_constraints, fact_type=AnnotationType.constraint,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_copyright_annotations, fact_type=AnnotationType.copyright,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_copyright, fact_type=AnnotationType.copyright,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_court_annotations, fact_type=AnnotationType.court,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=_get_courts, fact_type=AnnotationType.court,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_cusip_annotations, fact_type=AnnotationType.cusip,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_cusip, fact_type=AnnotationType.cusip,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_date_annotations, fact_type=AnnotationType.date,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_dates, fact_type=AnnotationType.date,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_definition_annotations, fact_type=AnnotationType.definition,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_definitions, fact_type=AnnotationType.definition,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_distance_annotations, fact_type=AnnotationType.distance,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_distances, fact_type=AnnotationType.distance,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_duration_annotations, fact_type=AnnotationType.duration,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_durations, fact_type=AnnotationType.duration,
                               result_fmt=ExtractorResultFormat.fmt_object),

            # SIC! dictionary is required
            ExtractingFunction(method=get_geoentity_annotations, fact_type=AnnotationType.geoentity,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_geoentities, fact_type=AnnotationType.geoentity,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_money_annotations, fact_type=AnnotationType.money,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_money, fact_type=AnnotationType.money,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_percent_annotations, fact_type=AnnotationType.percent,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_percents, fact_type=AnnotationType.percent,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_pii_annotations, fact_type=AnnotationType.pii,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_pii, fact_type=AnnotationType.pii,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_us_phone_annotations, fact_type=AnnotationType.phone,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_us_phones, fact_type=AnnotationType.phone,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_ssn_annotations, fact_type=AnnotationType.ssn,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_ssns, fact_type=AnnotationType.ssn,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_ratio_annotations, fact_type=AnnotationType.ratio,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_ratios, fact_type=AnnotationType.ratio,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_regulation_annotations, fact_type=AnnotationType.regulation,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_regulations, fact_type=AnnotationType.regulation,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_trademark_annotations, fact_type=AnnotationType.trademark,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_trademarks, fact_type=AnnotationType.trademark,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_url_annotations, fact_type=AnnotationType.url,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_urls, fact_type=AnnotationType.url,
                               result_fmt=ExtractorResultFormat.fmt_object),
        ]
        FactExtractor.store_functions(all_functs, FactExtractor.LANGUAGE_EN)

    @staticmethod
    def initialize_de():
        all_functs = [
            ExtractingFunction(method=get_de_amount_annotations, fact_type=AnnotationType.amount,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_de_amounts, fact_type=AnnotationType.amount,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_de_citation_annotations, fact_type=AnnotationType.citation,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_de_citations, fact_type=AnnotationType.citation,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_de_copyright_annotations, fact_type=AnnotationType.copyright,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_de_copyrights, fact_type=AnnotationType.copyright,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_de_court_citation_annotations, fact_type=AnnotationType.court_citation,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_de_court_citations, fact_type=AnnotationType.court_citation,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_de_court_annotations, fact_type=AnnotationType.court,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_de_courts, fact_type=AnnotationType.court,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_de_date_annotations, fact_type=AnnotationType.date,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_de_dates, fact_type=AnnotationType.date,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_de_definition_annotations, fact_type=AnnotationType.definition,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_de_definitions, fact_type=AnnotationType.definition,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_de_duration_annotations, fact_type=AnnotationType.duration,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_de_durations, fact_type=AnnotationType.duration,
                               result_fmt=ExtractorResultFormat.fmt_object),

            # SIC! dictionary is required
            ExtractingFunction(method=get_de_geoentity_annotations, fact_type=AnnotationType.geoentity,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_de_geoentities, fact_type=AnnotationType.geoentity,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_de_law_annotations, fact_type=AnnotationType.laws,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_de_laws, fact_type=AnnotationType.laws,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_de_percent_annotations, fact_type=AnnotationType.percent,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_de_percents, fact_type=AnnotationType.percent,
                               result_fmt=ExtractorResultFormat.fmt_object),
        ]
        FactExtractor.store_functions(all_functs, FactExtractor.LANGUAGE_DE)

    @staticmethod
    def initialize_es():
        all_functs = [
            ExtractingFunction(method=get_es_copyright_annotations, fact_type=AnnotationType.copyright,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_es_copyrights, fact_type=AnnotationType.copyright,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_es_court_annotations, fact_type=AnnotationType.court,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_es_courts, fact_type=AnnotationType.court,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_es_date_annotations, fact_type=AnnotationType.date,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_es_dates, fact_type=AnnotationType.date,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_es_definition_annotations, fact_type=AnnotationType.definition,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_es_definitions, fact_type=AnnotationType.definition,
                               result_fmt=ExtractorResultFormat.fmt_object),

            ExtractingFunction(method=get_es_regulation_annotations, fact_type=AnnotationType.regulation,
                               result_fmt=ExtractorResultFormat.fmt_class),
            ExtractingFunction(method=get_es_regulations, fact_type=AnnotationType.regulation,
                               result_fmt=ExtractorResultFormat.fmt_object),
        ]
        FactExtractor.store_functions(all_functs, FactExtractor.LANGUAGE_ES)

    @staticmethod
    def store_functions(all_functs: List[ExtractingFunction], lang: str):
        func_by_ret_fmt = {}  # { "class": { "amount": <routine>, ...
        for func in all_functs:
            func.lang = lang
            tp_func_dict = func_by_ret_fmt.get(func.result_fmt)
            if not tp_func_dict:
                tp_func_dict = {}
                func_by_ret_fmt[func.result_fmt] = tp_func_dict
            tp_func_dict[func.fact_type] = func

        FactExtractor.func_by_lang[lang] = func_by_ret_fmt


FactExtractor.initialize()



__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class Map(dict):
    """
    Example:
    m = Map(some_dict)
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    self[k] = v

        if kwargs:
            for k, v in kwargs.items():
                self[k] = v

        self.objectify(self)

    def objectify(self, a_dict):
        for key, val in a_dict.items():
            if isinstance(val, dict):
                a_dict[key] = Map(val)

    def __getattr__(self, attr):
        return self.get(attr)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.__dict__.update({key: value})

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        super().__delitem__(key)
        del self.__dict__[key]



from typing import Tuple, List

from lexnlp.extract.common.annotations.text_annotation import TextAnnotation

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class DefinitionAnnotation(TextAnnotation):
    record_type = 'definition'
    """
    create an object of DefinitionAnnotation like
    cp = DefinitionAnnotation(name='name', coords=(0, 100), text='text text')
    """
    def __init__(self,
                 coords: Tuple[int, int],
                 locale: str = 'en',
                 name: str = '',
                 text: str = None):
        super().__init__(
            name=name,
            locale=locale,
            coords=coords,
            text=text)

    def get_cite_value_parts(self) -> List[str]:
        return [self.name]

    # pylint: disable=unused-argument
    def get_extracted_text(self, full_text: str) -> str:
        return self.text

    def get_dictionary_values(self) -> dict:
        ant = {
            'tags': {
                'Extracted Entity Definition Name': self.name,
                'Extracted Entity Text': self.text or self.name
            }
        }
        return ant



__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class SpecialCharacters:
    punctuation = {'.', ',', ':', '-', ';', ')', '(', ']', '{', '}'
                   '[', '*', '/', '\\', '"', '\'', '!', '?', '%',
                   '$', '^', '&', '@'}



import re
from enum import Enum
from lexnlp.utils.lines_processing.line_processor import LineOrPhrase, LineProcessor

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


LineType = Enum('LineType', 'regular header paragraph_start')


class TypedLineOrPhrase(LineOrPhrase):
    """
    Extends the LineOrPhrase class (text - ending - position)
    Adds LineType attribute specifying the line's "role" within the text
    """
    def __init__(self):
        super().__init__()
        self.type = LineType.regular

    @staticmethod
    def wrap_line(l: LineOrPhrase):
        t = TypedLineOrPhrase()
        t.text = l.text
        t.start = l.start
        t.ending = l.ending
        return t

    def __repr__(self):
        return '[' + str(self.type) + '] ' + self.text + '->' + self.ending


class ParsedTextQualityEstimate:
    """
    A complex estimate for a text fragment
    - corrupted_prob: the probability of the text being "corrupted".
      Currently equals to extra_line_breaks_prob.
    - extra_line_breaks_prob: the probability (0..100) of the text containing unnecessary
      line breaks (\n).
    - an average line length withing the text (in characters).
    """
    def __init__(self):
        self.corrupted_prob = 0
        self.extra_line_breaks_prob = 0
        self.avg_line_length = 100


class ParsedTextQualityEstimator:
    """
    Estimates the probability of the text passed being somewhat corrupted
    """

    sentence_break_chars = {'.', ';', '!', '?', ','}
    reg_numered_header = re.compile(r'(^[\s]*\(?[a-zA-Z]\)?\s)|(^[\s]*[0-9\.]+[\)]?\s)')
    reg_paragraph_start = re.compile(r'(^\s{2})|(^\t)')
    minimal_paragraph_line_length = 250

    def __init__(self):
        self.estimate = ParsedTextQualityEstimate()
        self.lines = []
        self.proc = LineProcessor()

    def estimate_text(self, text: str) -> ParsedTextQualityEstimate:
        """
        Let's assume the text is:
            Contrary to popular belief, Lorem Ipsum is not simply random text. It has roots in a piece of classical
            Latin literature from 45 BC, making it over 2000 years old. Richard McClintock, a Latin professor at

            Hampden-Sydney College in Virginia, looked up one of the more obscure Latin words, consectetur, from a
            Lorem Ipsum passage, and going through the cites of the word in classical literature, discovered

            the undoubtable source.
        :param text: a text containing a number of \n\n sequences, see above
        :return: ParsedTextQualityEstimate: {'avg_line_length': 103, 'extra_line_breaks_prob': 66, 'corrupted_prob': 66}
        """

        self.split_text_on_lines(text)
        # does the text contain unnecessary line breaks?
        self.estimate_extra_line_breaks()
        # wrap up the estimate
        self.estimate.corrupted_prob = self.estimate.extra_line_breaks_prob
        return self.estimate

    def split_text_on_lines(self, text: str):
        self.estimate = ParsedTextQualityEstimate()

        self.lines = [TypedLineOrPhrase.wrap_line(l) for l in
                      self.proc.split_text_on_line_with_endings(text)]
        self.proc.determine_line_length(text)
        self.estimate.avg_line_length = self.proc.line_length

        for line in self.lines:
            self.determine_line_type(line)

    def estimate_extra_line_breaks(self):
        lines_total = len(self.lines)
        if lines_total == 0:
            return

        longest_seq = 0
        current_seq = 0
        total_extra_breaks = 0

        for indx in range(0, len(self.lines)):
            if self.check_line_followed_by_unnecessary_break(indx):
                total_extra_breaks += 1
                current_seq += 1
                longest_seq = max(current_seq, longest_seq)
                continue
            current_seq = 0

        if total_extra_breaks > 1:
            p1 = 100 if longest_seq > lines_total / 3 else int(100 * longest_seq * 2.5 / lines_total)
            p2 = int(100 * total_extra_breaks * 2 / lines_total)
            self.estimate.extra_line_breaks_prob = min(100, max(p1, p2))

    def check_line_followed_by_unnecessary_break(self, line_index: int) -> bool:
        line = self.lines[line_index]
        if line.ending.count('\n') <= 1:
            return False
        if len(line.text) > ParsedTextQualityEstimator.minimal_paragraph_line_length:
            # the whole line could be a paragraph
            return False
        prob_needs_extra = line.type == LineType.header
        if not prob_needs_extra:
            next_line = self.lines[line_index + 1] if line_index < len(self.lines) - 1 else None
            prob_needs_extra = next_line is not None and next_line.type != LineType.regular
        return not prob_needs_extra

    def determine_line_type(self, line: TypedLineOrPhrase):
        p_head = self.estimate_line_is_header_prob(line.text)
        if p_head > 50:
            line.type = LineType.header
            return
        p_par_start = self.estimate_line_is_paragraph_start_prob(line.text)
        if p_par_start > 50:
            line.type = LineType.paragraph_start

    def estimate_line_is_paragraph_start_prob(self, line: str) -> int:
        if ParsedTextQualityEstimator.reg_paragraph_start.search(line):
            return 100
        return 0

    def estimate_line_is_header_prob(self, line: str) -> int:
        line = line.rstrip(' \t')
        if len(line) == 0:
            return 0
        if line[-1] in ParsedTextQualityEstimator.sentence_break_chars:
            return 0
        if ParsedTextQualityEstimator.reg_numered_header.search(line):
            return 100

        if len(line) < self.estimate.avg_line_length * 0.6:
            return 65 # 65% chance the line is a header

        return 35



import re
from typing import List, Tuple

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


PhraseMatch = Tuple[str, int, int]


class PhraseFinder:
    """
    The class contains a collection of short string (usually 1 or 2 or 3 words)
    PhraseFinder searches for these strings (phrases) in the text given, either
    ignoring or regarding the case
    """

    def __init__(self, phrase_set: List[str], extra_format_function=None):
        self.extra_format_function = extra_format_function
        self.word_re_ig = dict((v, self.word_to_regex(v, True)) for v in phrase_set)
        self.word_re_cs = dict((v, self.word_to_regex(v, False)) for v in phrase_set)

    def word_to_regex(self, word: str, ignore_case: bool) -> str:
        # " Amtsgericht Stuttgart" ->  re("Amtsgericht[\s]+Stuttgart")
        subphrase = word.replace(r'\t', ' ').strip(' ').replace('  ', ' ').replace(' ', r'[\s]+')
        if self.extra_format_function is not None:
            subphrase = self.extra_format_function(subphrase)
        sps = '(\\b|\\s)'
        return re.compile(sps + subphrase + sps, re.IGNORECASE | re.UNICODE) if ignore_case else \
            re.compile(sps + subphrase + sps, re.UNICODE)

    def find_word(self, phrase: str, ignore_case: bool = True) -> List[PhraseMatch]:
        """
        :param phrase: "Tis better using France than trusting France: let us be back'd with God and with the seas"
        :param ignore_case: True
        :return: [ ('better', 4, 9), (' let us ', 46, 52) ]

        PhraseFinder instance had been initialized like
            PhraseFinder([' let us ', 'better', 'the sea'])
        """
        matches = []
        match_dict = self.word_re_ig if ignore_case else self.word_re_cs

        for k, v in match_dict.items():
            for match in v.finditer(phrase):
                matches.append((k, match.start(), match.end()))
        return matches



"""Regulation extraction for English.

This module implements regulation extraction functionality in English.

Todo:
  * Improved unit tests and case coverage
"""

# Imports
import regex as re

from typing import Generator

from lexnlp.extract.common.annotations.regulation_annotation import RegulationAnnotation

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


REGULATION_CODES_MAP = {
    'CFR': 'Code of Federal Regulations',
    'USC': 'United States Code',
}

REGULATION_PTN = r"""
(
(\d+)\s*
(U\.?S\.?C\.?|C\.?F\.?R\.?)\s*
(Sec(?:tion|\.)?|§)?\s*
(\d+[\da-zA-Z\-]*)
)"""
REGULATION_PTN_RE = re.compile(REGULATION_PTN, re.IGNORECASE | re.MULTILINE | re.DOTALL | re.VERBOSE)

PUBLIC_LAW_PTN = r"""
(
Pub(?:lic|\.)\s+L(?:aw|\.)(?:\s+No.?)?\s+\d+\-\d+
|
\d+\s+Stat\.\s+[\d-]+
)
"""
PUBLIC_LAW_PTN_RE = re.compile(PUBLIC_LAW_PTN, re.IGNORECASE | re.MULTILINE | re.DOTALL | re.VERBOSE)
PUBLIC_LAW_SUB_RE = re.compile(r'.+?(\d+\-\d+)', re.MULTILINE | re.DOTALL)
REGULATIONS_DICT_KEYS = ['regulation_type', 'regulation_code', 'regulation_str']


def get_regulations(text, return_source=False, as_dict=False) -> Generator:
    """
    Get regulations.
    :param text:
    :param return_source:
    :param as_dict:
    :return: tuple or dict
    (volume, reporter, reporter_full_name, page, page2, court, year[, source text])
    """
    for ant in get_regulation_annotations(text):
        if not as_dict:
            item = (ant.source, ant.name)
            if return_source:
                item += (ant.text,)
            yield item
        else:
            yield ant.to_dictionary_legacy()


def get_regulation_annotations(text: str) -> \
        Generator[RegulationAnnotation, None, None]:
    """
    Get regulations.
    :param text:
    :param return_source:
    :param as_dict:
    :return: tuple or dict
    (volume, reporter, reporter_full_name, page, page2, court, year[, source text])
    """

    for match in REGULATION_PTN_RE.finditer(text):
        source_text, num1, regulation_type, sec, num2 = match.groups()
        fixed_regulation_type = regulation_type.replace('.', '')
        if sec and 'sec' in sec.lower():
            sec = 'Section'
        regulation_parts = [num1, fixed_regulation_type, sec, num2] if sec else [num1, fixed_regulation_type, num2]
        item = (REGULATION_CODES_MAP.get(fixed_regulation_type, regulation_type),
                ' '.join(regulation_parts))

        ant = RegulationAnnotation(coords=match.span(),
                                   source=item[0],
                                   name=item[1],
                                   text=source_text.strip())
        yield ant

    for match in PUBLIC_LAW_PTN_RE.finditer(text):
        source_text = match.groups(0)
        if isinstance(source_text, tuple):
            source_text = source_text[0]
        if source_text and 'pub' in source_text.lower():
            fixed_code = PUBLIC_LAW_SUB_RE.sub(r'Public Law No. \1', source_text)
        else:
            fixed_code = ' '.join(source_text.lower().title().split())

        ant = RegulationAnnotation(coords=match.span(),
                                   source='Public Law',
                                   name=fixed_code,
                                   text=source_text.strip())
        yield ant



"""Urls extraction for English using NLTK and NLTK pre-trained maximum entropy classifier.

This module implements basic urls extraction functionality in English relying on the pre-trained
NLTK functionality, including POS tagger and NE (fuzzy) chunkers.

Todo: -
"""

# Imports
import re

from typing import Generator

from lexnlp.extract.common.annotations.url_annotation import UrlAnnotation

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


URL_PTN = r"""
(?xi)
\b
(                       # Capture 1: entire matched URL
  (?:
    https?://               # http or https protocol
    |                       #   or
    www\d{0,3}[.]           # "www.", "www1.", "www2." … "www999."
    |                           #   or
    [a-z0-9.\-]+[.][a-z]{2,4}/  # looks like domain name followed by a slash
  )
  (?:                       # One or more:
    [^\s()<>]+                  # Run of non-space, non-()<>
    |                           #   or
    \(([^\s()<>]+|(\([^\s()<>]+\)))*\)  # balanced parens, up to 2 levels
  )+
  (?:                       # End with:
    \(([^\s()<>]+|(\([^\s()<>]+\)))*\)  # balanced parens, up to 2 levels
    |                               #   or
    [^\s`!()\[\]{};:'".,<>?«»“”‘’]        # not a space or one of these punct chars
  )
)
"""
URL_PTN_RE = re.compile(URL_PTN, re.IGNORECASE | re.MULTILINE | re.VERBOSE)


def get_urls(text: str) -> Generator[str, None, None]:
    """
    Find urls in text.
    """
    for ant in get_url_annotations(text):
        yield ant.url


def get_url_annotations(text: str) -> Generator[UrlAnnotation, None, None]:
    """
    Find urls in text.
    """
    for match in URL_PTN_RE.finditer(text):
        url = match.group()
        ant = UrlAnnotation(coords=match.span(),
                            url=url)
        yield ant



from collections import Iterable

from typing import Callable, Any

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


def collapse_sequence(sequence: Iterable,
                      predicate: Callable[[Any, Any], Any],
                      accumulator: Any = 0.0) -> Any:
    for item in sequence:
        accumulator = predicate(item, accumulator)
    return accumulator


def count_sequence_matches(sequence: Iterable,
                           predicate: Callable[[Any], bool]) -> int:
    return collapse_sequence(sequence,
                             lambda i, a: a + 1 if predicate(i) else a, 0)



"""Citation extraction for English.

This module implements citation extraction functionality in English.

Todo:
  * Improved unit tests and case coverage
"""

# pylint: disable=bare-except

# Imports
from typing import Generator

import regex as re
from reporters_db import EDITIONS, REPORTERS

from lexnlp.extract.common.annotations.citation_annotation import CitationAnnotation

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


CITATION_PTN = r"""
(?:[\s,:\(]|^)
(
(\d+)\s+
({reporters})\s+
(\d+)
(?:,\s+(\d+(?:\-\d+)?))?
(?:\s+\((.+?)?(\d{{4}})\))?
)
(?:\W|$)
""".format(reporters='|'.join([re.escape(i) for i in EDITIONS]))
CITATION_PTN_RE = re.compile(CITATION_PTN, re.IGNORECASE | re.MULTILINE | re.DOTALL | re.VERBOSE)


def get_citations(text: str, return_source=False, as_dict=False) -> Generator:
    """
    Get citations.
    :param text:
    :param return_source:
    :param as_dict:
    :return: tuple or dict
    (volume, reporter, reporter_full_name, page, page2, court, year[, source text])
    """
    for ant in get_citation_annotations(text):
        if as_dict:
            yield ant.to_dictionary_legacy()
        else:
            item = (ant.volume,
                    ant.reporter,
                    ant.reporter_full_name,
                    ant.page,
                    ant.page_range,
                    ant.court,
                    ant.year)
            if return_source:
                item += (ant.source,)
            yield item


def get_citation_annotations(text: str) -> \
        Generator[CitationAnnotation, None, None]:
    """
    Get citations.
    :param text:
    :param return_source:
    :param as_dict:
    :return: tuple or dict
    (volume, reporter, reporter_full_name, page, page2, court, year[, source text])
    """
    for match in CITATION_PTN_RE.finditer(text):
        source_text, volume, reporter, \
            page, page2, court, year = match.groups()
        try:
            reporter_data = REPORTERS[EDITIONS[reporter]]
            reporter_full_name = ''
            if len(reporter_data) == 1:
                reporter_full_name = reporter_data[0]['name']
            elif year:
                for period_data in reporter_data:
                    if reporter in period_data['editions']:
                        start = period_data['editions'][reporter]['start'].year
                        end = period_data['editions'][reporter]['end']
                        if (end and start <= int(year) <= end.year) or start <= int(year):
                            reporter_full_name = period_data['name']

            ant = CitationAnnotation(coords=match.span(),
                                     volume=int(volume) if volume else None,
                                     year=int(year) if year and year.isdigit() else None,
                                     reporter=reporter,
                                     reporter_full_name=reporter_full_name,
                                     page=int(page) if page else None,
                                     page_range=page2,
                                     source=source_text.strip(),
                                     court=court.strip(', ') if court else None,
                                     locale='en')

            yield ant
        except KeyError:
            pass



import gzip
import os
import pickle
from abc import abstractmethod
from typing import Any, Tuple, Generator, List

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


MODULE_PATH = os.path.abspath(os.path.dirname(__file__))


class BaseTokenSequenceClassifierModel(object):
    """
    Base classifier class for generic text sequence objects.
    """

    unicode_character_categories = pickle.load(
        open(os.path.join(MODULE_PATH, 'data/unicode_character_categories.pickle'), 'rb'))
    unicode_character_category_mapping = pickle.load(
        open(os.path.join(MODULE_PATH, 'data/unicode_character_category_mapping.pickle'), 'rb'))
    unicode_character_top_category_mapping = pickle.load(
        open(os.path.join(MODULE_PATH, 'data/unicode_character_top_category_mapping.pickle'), 'rb'))
    unicode_top_category_set = set(unicode_character_top_category_mapping.values())
    unicode_category_set = set(unicode_character_category_mapping.values())

    @staticmethod
    # pylint: disable=unused-argument
    def get_classifier(
            use_spacy: bool,
            letter_set=None, digit_set=None, punc_set=None, symbol_set=None, match_tokens=None,
            pre_window=0, post_window=0, calculate_sum=False, normalize=False, string_checks=False):
        init_args = dict(locals())
        del init_args['use_spacy']
        from lexnlp.extract.ml.classifier.spacy_token_sequence_model import SpacyTokenSequenceClassifierModel
        from lexnlp.extract.ml.classifier.token_sequence_model import TokenSequenceClassifierModel
        return SpacyTokenSequenceClassifierModel(**init_args) if use_spacy \
            else TokenSequenceClassifierModel(**init_args)

    # pylint: disable=unused-argument
    def __init__(self, letter_set=None, digit_set=None, punc_set=None, symbol_set=None, match_tokens=None,
                 pre_window=0, post_window=0, calculate_sum=False, normalize=False, string_checks=False):
        # store init arguments for saving model
        init_args = dict(locals())
        del init_args['self']
        self.init_args = (type(self), init_args)
        # Classifier parameters
        self.letter_set = letter_set if letter_set else []
        self.digit_set = digit_set if digit_set else []
        self.punc_set = punc_set if punc_set else []
        self.symbol_set = symbol_set if symbol_set else []
        self.match_tokens = match_tokens if match_tokens else []
        self.pre_window = pre_window
        self.post_window = post_window
        self.calculate_sum = calculate_sum
        self.normalize = normalize
        self.string_checks = string_checks

        # "Private" class variables
        self.model = None
        self.feature_list = self.get_feature_list(self.letter_set, self.digit_set,
                                                  self.punc_set, self.symbol_set,
                                                  self.pre_window, self.post_window)
        self._feature_index_map = dict([(f, i) for i, f in enumerate(self.feature_list)])
        self._base_feature_list = [f[2:] for f in self.feature_list if f.startswith('0_')]

    def save_in_file(self, save_path: str):
        with open(save_path, 'wb') as fw:
            pickle.dump(self, fw)

    def save_in_file_compressed(self, save_path: str):
        with gzip.GzipFile(save_path, 'w') as fw:
            pickle.dump(self, fw)

    @staticmethod
    def load_from_file(save_path: str):
        with open(save_path, 'rb') as fr:
            model = pickle.load(fr)
        return model

    @staticmethod
    def load_from_file_compressed(save_path: str):
        with gzip.GzipFile(save_path, 'r') as fr:
            model = pickle.load(fr)
        return model

    @staticmethod
    def load_from_stream(stream: Any):
        model = pickle.load(stream)
        return model

    @abstractmethod
    def get_feature_list(self, letter_set=None, digit_set=None, punc_set=None, symbol_set=None,
                         pre_window=None, post_window=None):
        raise NotImplementedError('get_feature_list() should be implemented in derived class')

    @abstractmethod
    def get_feature_data(self, text: str, feature_mask: List[int] = None):
        raise NotImplementedError('get_feature_data() should be implemented in derived class')

    def train_model(self, model, feature_data, target_data):
        """
        Train a model and set into class.
        """
        self.model = model.fit(feature_data, target_data)

    def run_model(self, text: str, outer_class=0,
                  start_class=1, inner_class=2, end_class=3, strict=True,
                  feature_mask: List[int] = None)\
            -> Generator[Tuple[int, int], None, None]:
        """
        Run model on text
        """

        feature_data, tokens = self.get_feature_data(text, feature_mask)
        predicted_class = self.model.predict(feature_data)
        start_pos = -1

        for i in range(0, predicted_class.shape[0]):
            if predicted_class[i] == start_class and start_pos == -1:
                start_pos = i
            elif predicted_class[i] == end_class:
                if strict:
                    if start_pos >= 0:
                        yield (tokens[start_pos][0], tokens[i][1])
                        start_pos = -1
                else:
                    yield (tokens[start_pos][0], tokens[i][1])
                    start_pos = -1
            elif predicted_class[i] == inner_class and start_pos == -1:
                start_pos = i
            elif predicted_class[i] == outer_class and start_pos > -1:
                if not strict:
                    yield (tokens[start_pos][0], tokens[i][1])
                start_pos = -1



# pylint: disable=unused-import

from typing import Pattern, List, Generator, Tuple
# pylint: enable=unused-import
from lexnlp.extract.common.copyrights.copyright_en_style_parser import CopyrightEnStyleParser
from lexnlp.extract.common.annotations.copyright_annotation import CopyrightAnnotation
from lexnlp.extract.es.language_tokens import EsLanguageTokens
from lexnlp.utils.lines_processing.line_processor import LineSplitParams
from lexnlp.utils.lines_processing.line_processor import LineProcessor

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class CopyrightEsParser(CopyrightEnStyleParser):
    line_processor = None  # type:LineProcessor

    @staticmethod
    def init_parser():
        split_params = LineSplitParams()
        split_params.line_breaks = {'\n', '.', ';', '!', '?'}
        split_params.abbreviations = EsLanguageTokens.abbreviations
        split_params.abbr_ignore_case = True
        CopyrightEsParser.line_processor = LineProcessor(
            line_split_params=split_params)

    @classmethod
    def extract_phrases_with_coords(cls, sentence: str) -> List[Tuple[str, int, int]]:
        return [(t.text, t.start, t.get_end()) for t in
                cls.line_processor.split_text_on_line_with_endings(sentence)]


CopyrightEsParser.init_parser()


def get_copyright_annotations(text: str, return_sources=False) -> \
        Generator[CopyrightAnnotation, None, None]:
    for ant in  CopyrightEsParser.get_copyright_annotations(text,
                                                            return_sources):
        ant.locale = 'es'
        yield ant


def get_copyrights(text: str, return_sources=False) -> \
        Generator[dict, None, None]:
    for ant in get_copyright_annotations(text, return_sources):
        yield ant.to_dictionary()


def get_copyright_list(text: str, return_sources=False) -> List[CopyrightAnnotation]:
    return list(get_copyright_annotations(text, return_sources))



from itertools import groupby
from typing import Callable, List
from lexnlp.extract.common.annotations.text_annotation import TextAnnotation
from lexnlp.extract.common.pattern_found import PatternFound
from lexnlp.utils.lines_processing.line_processor import LineProcessor, LineSplitParams, LineOrPhrase

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class TextPatternCollector:
    basic_line_processor = LineProcessor()
    """
    EsDefinitionsParser searches for definitions in text according to the
    rules of Spanish. See the "parse" method
    """
    def __init__(self, parsing_functions: List[Callable[[str], List[PatternFound]]],
                 split_params: LineSplitParams):
        """
        :param parsing_functions: a functions' collection from SpanishParsingMethods
        :param split_params: text-to-sentences splitting params
        """
        self.parsing_functions = parsing_functions
        self.annotations = [] # type: List[TextAnnotation]
        self.split_params = split_params
        self.proc = LineProcessor(line_split_params=self.split_params)
        self.prohibited_words = {} # words that are Not definitions per se

    def parse(self, text: str, locale: str = None) -> List[TextAnnotation]:
        """
        :param locale: 'En', 'De', 'Es', ...
        :param text: En este acuerdo, el término "Software" se refiere a: (i) el programa informático
        :return: { "attrs": {"start": 28, "end": 82}, "tags": {"Extracted Entity Type": "definition",
                "Extracted Entity Definition Name": "Software",
                "Extracted Entity Text": ""Software" se refiere a: (i) el programa informático"} }
        """
        self.annotations = []  # type: List[TextAnnotation]
        for phrase in self.proc.split_text_on_line_with_endings(text):
            matches = []
            for f in self.parsing_functions:
                ml = f(phrase.text)
                matches += ml
            # find synonyms
            # sort and take the most appropriate matches
            matches = self.remove_prohibited_words(matches)
            if len(matches) > 1:
                matches = self.choose_best_matches(matches)
                matches = self.choose_more_precise_matches(matches, phrase.text)
            # trim parts of matches
            for match in matches:
                # make_annotation_from_pattrn DOES return data in derived classes
                # pylint:disable=assignment-from-none
                ant = self.make_annotation_from_pattrn(locale, match, phrase)
                ant.coords = (ant.coords[0] + phrase.start, ant.coords[1] + phrase.start)
                self.annotations.append(ant)
        return self.annotations

    # pylint: disable=unused-argument
    def make_annotation_from_pattrn(self, locale: str,
                                    ptrn: PatternFound,
                                    phrase: LineOrPhrase) -> TextAnnotation:
        # should be overriden in derived class
        return None

    # pylint: enable=unused-argument

    def remove_prohibited_words(self, matches: List[PatternFound]) -> List[PatternFound]:
        # like 'und' or 'and' or 'the' - the word like this is not a definition itself
        return [m for m in matches if m.name not in self.prohibited_words]

    def choose_best_matches(self, matches: List[PatternFound]) -> List[PatternFound]:
        resulted = []
        # pylint: disable=unused-variable
        for _, g in groupby(matches, lambda m: m.name.strip(" \t'\"")):
        # pylint: enable=unused-variable
            same_matches = list(g)
            if len(same_matches) > 1:
                same_matches = [sorted(same_matches,
                                       key=TextPatternCollector.estimate_match_quality, reverse=True)[0]]
            resulted += same_matches
        return resulted

    def choose_more_precise_matches(self,
                                    matches: List[PatternFound],
                                    text: str) -> List[PatternFound]:
        """
        look for a match "consumed" by other matches and spare the consuming! matches
        """
        resulted = []
        if len(matches) < 2:
            return matches
        for i in range(0, len(matches)):
            a = matches[i]
            a_worse_b = False
            for j in range(0, len(matches)):
                if i == j:
                    continue
                b = matches[j]
                if a.pattern_worse_than_target(b, text):
                    a_worse_b = True
                    break
            if not a_worse_b:
                resulted.append(a)
        return resulted

    @staticmethod
    def estimate_match_quality(match: PatternFound) -> int:
        return 1000 * match.probability - (match.end - match.start)



"""Distance extraction for English.

This module implements basic distance extraction functionality in English.

"""

# Imports
import re
from typing import Generator

from lexnlp.extract.common.annotations.distance_annotation import DistanceAnnotation
from lexnlp.extract.en.amounts import get_amounts, NUM_PTN

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


DISTANCE_SYMBOL_MAP = {
    "km": "kilometer",
    "mi": "mile",
}

DISTANCE_TOKEN_MAP = {
    "kilometers": "kilometer",
    "kilometer": "kilometer",
    "miles": "mile",
    "mile": "mile",
}

DISTANCE_PTN = r"""
(({num_ptn})\s*
({distance_tokens}|{distance_symbols}))(?:\W|$)
""".format(
    num_ptn=NUM_PTN.replace('(?:\\W|$)', '').replace('(?<=\\W|^)', ''),
    distance_symbols='|'.join(DISTANCE_SYMBOL_MAP),
    distance_tokens='|'.join(DISTANCE_TOKEN_MAP)
)
DISTANCE_PTN_RE = re.compile(DISTANCE_PTN, re.IGNORECASE | re.MULTILINE | re.DOTALL | re.VERBOSE)


def get_distances(text: str, return_sources=False, float_digits=4) -> Generator:
    for ant in get_distance_annotations(text, float_digits):
        item = (ant.amount, ant.distance_type)
        if return_sources:
            item += (ant.text,)
        yield item


def get_distance_annotations(text: str, float_digits=4) \
        -> Generator[DistanceAnnotation, None, None]:

    for match in DISTANCE_PTN_RE.finditer(text.lower()):
        source_text, number_text, distance_item = match.groups()
        amount = list(get_amounts(number_text, float_digits=float_digits))
        if len(amount) != 1:
            continue
        distance_type = DISTANCE_SYMBOL_MAP.get(distance_item) \
                        or DISTANCE_TOKEN_MAP.get(distance_item)
        amount = amount[0]
        if float_digits:
            amount = round(amount, float_digits)
        ant = DistanceAnnotation(coords=match.span(),
                                 amount=amount,
                                 distance_type=distance_type,
                                 text=source_text.strip())
        yield ant



"""Citation extraction for German.

This module implements citation extraction functionality in German.

Todo:
  * Improved unit tests and case coverage
"""

# pylint: disable=bare-except

import regex as re
from typing import Generator

from lexnlp.extract.common.annotations.text_annotation import TextAnnotation
from lexnlp.extract.common.annotations.citation_annotation import CitationAnnotation
from lexnlp.extract.en.dates import get_dates

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class DeCitationParser:
    CITATION_PTN = r"""
    (?P<text>
    (?:
    (?:
    (?:
    (?:Artikel\s(?P<article>\d+))|
    (?:\sSatz\s(?P<sentence>\d+))|
    (?:\s(?:Nummer|Nr\.)\s(?P<number>(?:\d+|\s*und |\s*bis\s)+))|
    (?:\sAbsatz(?:es)?\s(?P<paragraph>(?:\d+|\s*und\s)+))|
    (?:\sUnterabsatz(?:es)?\s(?P<subparagraph>(?:\d+|\s*und\s)+))|
    (?:\sBuchstabe\s(?P<letter>(?:\w|\s*und\s)+))
    )*
    (?:\sdes\sGesetzes|\sder\sVerordnung)
    )?\s*
    (?:vom|v\.)\s(?P<date>\d{1,2}\.\s\S+\s\d{4}),?\s
    )?
    \(?BGBl\.?\s(?:(?P<year>\d{4})\s)?(?P<part>I{1,3})\sS\.\s(?:(?P<page>\d+)(?:\s\((?P<year>\d{4})\))?(?:,\s)?)+\)?)
    """

    SECOND_CITATION_PTN = r"""
    (?P<text>BGBl[IS\.,\s\d]+;\s(?P<year>\d{4})\s(?P<part>I{1,3})\sS\.\s(?:(?P<page>\d+)(?:,\s)?)+)
    """

    CITATION_RANGE_PTN = r"""
    (?P<text>BGBl\.?\s(?P<part>I{1,3})\s(?P<year>\d{4}),\s(?P<page>\d+\s*[—-]\s*\d+))
    """

    re_flags = re.IGNORECASE | re.MULTILINE | re.DOTALL | re.VERBOSE
    CITATION_PTN_RE = re.compile(CITATION_PTN, re_flags)
    SECOND_CITATION_PTN_RE = re.compile(SECOND_CITATION_PTN, re_flags)
    CITATION_RANGE_PTN_RE = re.compile(CITATION_RANGE_PTN, re_flags)

    @classmethod
    def get_citation_annotations(cls, text: str) -> \
            Generator[CitationAnnotation, None, None]:
        """
        Get citations containing "BGBl"
        :param text: str
        :return: yields dict
        """

        for ptn in [cls.CITATION_PTN_RE, cls.SECOND_CITATION_PTN_RE, cls.CITATION_RANGE_PTN_RE]:
            for match in ptn.finditer(text):
                capture = match.capturesdict()
                date = ''.join(capture.get('date', ''))
                if date:
                    try:
                        date = str(list(get_dates(date, 'de'))[0]['value'])
                    except:
                        pass

                ant = CitationAnnotation(coords=match.span(),
                                         text=capture['text'][0],
                                         paragraph=''.join(capture.get('paragraph', '')),
                                         subparagraph=''.join(capture.get('subparagraph', '')),
                                         letter=''.join(capture.get('letter', '')),
                                         date=date,
                                         part=capture['part'][0],
                                         locale='de')
                ant.article = TextAnnotation.get_int_value(
                    ''.join(capture.get('article', '')))
                ant.number = TextAnnotation.get_int_value(
                    ''.join(capture.get('number', '')))
                ant.sentence = TextAnnotation.get_int_value(
                    ''.join(capture.get('sentence', '')))

                page_range = ', '.join(capture['page'])
                page = TextAnnotation.get_int_value(page_range)
                if page:
                    ant.page = page
                else:
                    ant.page_range = page_range

                volume_str = ''.join(capture.get('number', ''))
                volume = TextAnnotation.get_int_value(volume_str)
                if volume:
                    ant.volume = volume
                else:
                    ant.volume_str = volume_str

                year_str = ', '.join(capture.get('year', ''))
                year = TextAnnotation.get_int_value(year_str)
                if year:
                    ant.year = year
                else:
                    ant.year_str = year_str

                yield ant


def get_citation_annotations(text: str) -> \
        Generator[CitationAnnotation, None, None]:
    yield from DeCitationParser.get_citation_annotations(text)


def get_citations(text: str):
    """
    Get citations containing "BGBl"
    :param text: str
    :return: yields dict
    """
    for ant in get_citation_annotations(text):
        yield dict(
            location_start=ant.coords[0],
            location_end=ant.coords[1],
            text=ant.text,
            article=str(ant.article) if ant.article else '',
            number=str(ant.volume) if ant.volume else ant.volume_str,
            sentence=str(ant.sentence) if ant.sentence else '',
            date=ant.date,
            page=str(ant.page) if ant.page else ant.page_range,
            year=str(ant.year) if ant.year else ant.year_str,
            letter=ant.letter,
            paragraph = ant.paragraph,
            subparagraph = ant.subparagraph,
            part=ant.part
        )


def get_citation_list(text):
    return list(get_citations(text))



# Imports

from typing import List, Optional

import numpy

from lexnlp.extract.ml.classifier.base_token_sequence_classifier_model import BaseTokenSequenceClassifierModel

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class TokenSequenceClassifierModel(BaseTokenSequenceClassifierModel):
    """
    Classifier for generic text sequence objects.
    TODO: Implement skipgram.
    TODO: Implement sum and normalize.
    """

    def __init__(self, letter_set=None, digit_set=None, punc_set=None, symbol_set=None, match_tokens=None,
                 pre_window=0, post_window=0, calculate_sum=False, normalize=False, string_checks=False):
        super().__init__(letter_set=letter_set, digit_set=digit_set, punc_set=punc_set,
                         symbol_set=symbol_set, match_tokens=match_tokens, pre_window=pre_window,
                         post_window=post_window, calculate_sum=calculate_sum,
                         normalize=normalize, string_checks=string_checks)

    def get_feature_list(self,
                         letter_set=None,
                         digit_set=None,
                         punc_set=None,
                         symbol_set=None,
                         pre_window: Optional = None,
                         post_window: Optional = None):
        """
        Return set of features for a character tokenizer.
        """
        # handle params
        if letter_set is None:
            letter_set = self.letter_set

        if digit_set is None:
            digit_set = self.digit_set

        if punc_set is None:
            punc_set = self.punc_set

        if symbol_set is None:
            symbol_set = self.symbol_set

        if pre_window is None:
            pre_window = self.pre_window

        if post_window is None:
            post_window = self.post_window

        # initialize
        token_feature_list = ['position',
                              'length',
                              'mask']

        # iterate across window positions
        # pylint: disable=invalid-unary-operand-type
        for i in range(-pre_window, post_window + 1):
            o = str(i)
            token_feature_list.append(o + "_is_start")
            token_feature_list.append(o + "_is_end")
            if self.string_checks:
                token_feature_list.append(o + "_is_title")
                token_feature_list.append(o + "_is_lower")
                token_feature_list.append(o + "_is_upper")

            for c in letter_set:
                token_feature_list.append(o + "_char_" + c)
                token_feature_list.append(o + "_lchar_" + c)
                token_feature_list.append(o + "_first_char_" + c)
                token_feature_list.append(o + "_first_lchar_" + c)
                token_feature_list.append(o + "_last_char_" + c)
                token_feature_list.append(o + "_last_lchar_" + c)

            for c in digit_set:
                token_feature_list.append(o + "_digit_" + c)
                token_feature_list.append(o + "_first_digit_" + c)
                token_feature_list.append(o + "_last_digit_" + c)

            for c in punc_set:
                token_feature_list.append(o + "_punc_" + c)
                token_feature_list.append(o + "_first_punc_" + c)
                token_feature_list.append(o + "_last_punc_" + c)

            for c in symbol_set:
                token_feature_list.append(o + "_symbol_" + c)
                token_feature_list.append(o + "_first_symbol_" + c)
                token_feature_list.append(o + "_last_symbol_" + c)

            token_feature_list.append(o + "_char_other")
            token_feature_list.append(o + "_first_char_other")
            token_feature_list.append(o + "_last_char_other")

            for c in self.unicode_category_set:
                token_feature_list.append(o + "_cat_" + c)
                token_feature_list.append(o + "_first_cat_" + c)
                token_feature_list.append(o + "_last_cat_" + c)
            for c in self.unicode_top_category_set:
                token_feature_list.append(o + "_tcat_" + c)
                token_feature_list.append(o + "_first_tcat_" + c)
                token_feature_list.append(o + "_last_tcat_" + c)

            for token in self.match_tokens:
                token_feature_list.append(o + "_token_" + token)

        if self.calculate_sum:
            token_feature_list.append("sum_char_other")

            for c in self.letter_set:
                token_feature_list.append("sum_char_" + c)

            for c in self.digit_set:
                token_feature_list.append("sum_digit_" + c)

            for c in self.punc_set:
                token_feature_list.append("sum_punc_" + c)

            for c in self.symbol_set:
                token_feature_list.append("sum_symbol_" + c)

        return token_feature_list

    # TODO: add "hints" alongside with the text? areas with special
    def get_feature_data(self,
                         text: str,
                         feature_mask: List[int] = None):
        """
        Get features based on character model.
        feature_mask - array of numbers, has the same length as text
        """
        # setup return data
        text_len = len(text)
        text_lower = [c.lower() for c in text]
        text_data = [(self.unicode_character_top_category_mapping[
                          c] if c in self.unicode_character_top_category_mapping else "C",
                      self.unicode_character_category_mapping[
                          c] if c in self.unicode_character_category_mapping else "Cc",
                      ord(c)
                      ) for c in text]

        tokens = []
        token_start = 0
        # calculate token offsets
        for i in range(text_len - 1):
            if text_data[i][0] not in ['Z', 'C'] and text_data[i + 1][0] in ['Z', 'C']:
                tokens.append((token_start, i + 1))
                token_start = min(i + 2, text_len)
            elif text_data[i][0] in ['Z', 'C'] and text_data[i + 1][0] in ['Z', 'C']:
                token_start += 1
        tokens.append((token_start, text_len))
        num_tokens = len(tokens)

        # Setup return structure
        feature_data = numpy.zeros((num_tokens, len(self.feature_list)), dtype=numpy.int8)

        for i in range(num_tokens):
            o = str(0)
            token_text = text[tokens[i][0]:tokens[i][1]]

            feature_data[i, self._feature_index_map['position']] = i
            feature_data[i, self._feature_index_map['length']] = tokens[i][1] - tokens[i][0]
            feature_data[i, self._feature_index_map['mask']] = 0

            feature_data[i, self._feature_index_map[o + "_is_start"]] = int(i == 0)
            feature_data[i, self._feature_index_map[o + "_is_end"]] = int(i == num_tokens - 1)
            if self.string_checks:
                feature_data[i, self._feature_index_map[o + "_is_title"]] = int(token_text == token_text.title())
                feature_data[i, self._feature_index_map[o + "_is_lower"]] = int(token_text == token_text.lower())
                feature_data[i, self._feature_index_map[o + "_is_upper"]] = int(token_text == token_text.upper())

            if token_text in self.match_tokens:
                feature_data[i, self._feature_index_map[o + "_token_" + token_text]] = 1

            for j in range(tokens[i][0], tokens[i][1]):
                c = text[j]
                cl = text_lower[j]
                if feature_mask:
                    feature_data[i, self._feature_index_map['mask']] = max(
                        feature_data[i, self._feature_index_map['mask']], feature_mask[j])

                if c in self.letter_set:
                    feature_data[i, self._feature_index_map[o + "_char_" + c]] += 1
                    feature_data[i, self._feature_index_map[o + "_lchar_" + cl]] += 1
                    if j == tokens[i][0]:
                        feature_data[i, self._feature_index_map[o + "_first_char_" + c]] += 1
                        feature_data[i, self._feature_index_map[o + "_first_lchar_" + cl]] += 1
                    if j == tokens[i][1] - 1:
                        feature_data[i, self._feature_index_map[o + "_last_char_" + c]] += 1
                        feature_data[i, self._feature_index_map[o + "_last_lchar_" + cl]] += 1
                elif c in self.digit_set:
                    feature_data[i, self._feature_index_map[o + "_digit_" + c]] += 1
                    if j == tokens[i][0]:
                        feature_data[i, self._feature_index_map[o + "_first_digit_" + c]] += 1
                    if j == tokens[i][1] - 1:
                        feature_data[i, self._feature_index_map[o + "_last_digit_" + c]] += 1
                elif c in self.punc_set:
                    feature_data[i, self._feature_index_map[o + "_punc_" + c]] += 1
                    if j == tokens[i][0]:
                        feature_data[i, self._feature_index_map[o + "_first_punc_" + c]] += 1
                    if j == tokens[i][1] - 1:
                        feature_data[i, self._feature_index_map[o + "_last_punc_" + c]] += 1
                elif c in self.symbol_set:
                    feature_data[i, self._feature_index_map[o + "_symbol_" + c]] += 1
                    if j == tokens[i][0]:
                        feature_data[i, self._feature_index_map[o + "_first_symbol_" + c]] += 1
                    if j == tokens[i][1] - 1:
                        feature_data[i, self._feature_index_map[o + "_last_symbol_" + c]] += 1
                else:
                    feature_data[i, self._feature_index_map[o + "_char_other"]] += 1
                    if j == tokens[i][0]:
                        feature_data[i, self._feature_index_map[o + "_first_char_other"]] += 1
                    if j == tokens[i][1] - 1:
                        feature_data[i, self._feature_index_map[o + "_last_char_other"]] += 1

                feature_data[i, self._feature_index_map[o + "_cat_" + text_data[j][1]]] += 1
                feature_data[i, self._feature_index_map[o + "_tcat_" + text_data[j][0]]] += 1

                if j == tokens[i][0]:
                    feature_data[i, self._feature_index_map[o + "_first_cat_" + text_data[j][1]]] += 1
                    feature_data[i, self._feature_index_map[o + "_first_tcat_" + text_data[j][0]]] += 1
                if j == tokens[i][1] - 1:
                    feature_data[i, self._feature_index_map[o + "_last_cat_" + text_data[j][1]]] += 1
                    feature_data[i, self._feature_index_map[o + "_last_tcat_" + text_data[j][0]]] += 1

        # handle window feature calculations
        if self.pre_window + self.post_window > 0:
            for i in range(num_tokens):
                j_start = (i - self.pre_window) if (i - self.pre_window >= 0) else 0
                j_end = (i + self.post_window) if (i + self.post_window + 1 < num_tokens) else num_tokens
                for j in range(j_start, j_end):
                    if j == i:
                        continue
                    o = str(j - i)
                    for f in self._base_feature_list:
                        feature_data[i, self._feature_index_map[o + '_' + f]] = \
                            feature_data[j, self._feature_index_map['0_' + f]]

        return feature_data, tokens



"""Definition extraction for English.

This module implements basic definition extraction functionality in English.

Todo:
  * Improved unit tests and case coverage
"""

# pylint: disable=broad-except,bare-except


# Imports
import regex as re
import unidecode as unidecode
from collections import Counter
from typing import Pattern, List, Tuple, Set

from lexnlp.extract.common.annotations.phrase_position_finder import PhrasePositionFinder
from lexnlp.extract.common.text_beautifier import TextBeautifier
from lexnlp.extract.en.introductory_words_detector import IntroductoryWordsDetector
from lexnlp.extract.en.preprocessing.span_tokenizer import SpanTokenizer
from lexnlp.extract.common.special_characters import SpecialCharacters
from lexnlp.extract.en.en_language_tokens import EnLanguageTokens
from lexnlp.utils.lines_processing.line_processor import LineProcessor
from lexnlp.utils.iterating_helpers import count_sequence_matches

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class DefinitionCaught:
    """
    Each definition is stored in this class with
    its name, full text and "coords" within the whole document
    """
    __slots__ = ['name', 'text', 'coords']

    def __init__(self, name: str, text: str, coords: Tuple[int, int]):
        self.name = name
        self.text = text
        self.coords = coords

    def __repr__(self):
        return '%s [%d, %d]' % (self.name, self.coords[0], self.coords[1])

    def does_consume_target(self, target) -> int:
        """
        :param target: a definition that is, probably, "consumed" by the current one
        :return: 1 if self consumes the target, -1 if the target consumes self, overwise 0
        """
        coords_spans = (target.coords[0] >= self.coords[0] and
            target.coords[0] <= self.coords[1]) or \
            (self.coords[0] >= target.coords[0] and
            self.coords[0] <= target.coords[1])
        if not coords_spans:
            return 0
        if target.name in self.name:
            return 1
        if self.name in target.name:
            return -1
        return 0


# when the following flag is True,
# the term defined got stripped off quotes:
# e.g., (any such excess being referred to as a "Combined EDITT Deficit Alpha Beta Gamma Cappa")
# will become just (Combined EDITT Deficit Alpha Beta Gamma Cappa)
PICK_DEFINITION_FROM_QUOTES = True

# Constraints for matching
MAX_TERM_TOKENS = 5
MAX_QUOTED_TERM_TOKENS = 7
MAX_TERM_CHARS = 64

# Primary pattern triggers
STRONG_TRIGGER_LIST = ["shall have the meaning", r"includes?", "as including",
                       "shall mean", r"means?", r"shall (?:not\s+)?include",
                       "shall for purposes", "have meaning",
                       "referred to", "known as",
                       "refers to", "shall refer to", "as used",
                       "for purpose[sd]", "shall be deemed to",
                       "may be used", "is hereby changed to",
                       "is defined", "shall be interpreted"]

WEAK_TRIGGER_LIST = [r"[\(\)]", "in "]
ALL_TRIGGER_LIST = STRONG_TRIGGER_LIST + WEAK_TRIGGER_LIST

STRONG_TRIGGER_LIST.sort(key=len, reverse=True)
WEAK_TRIGGER_LIST.sort(key=len, reverse=True)
ALL_TRIGGER_LIST.sort(key=len, reverse=True)


def join_collection(collection):
    return "|".join([w.replace(" ", r"\s+") for w in collection])


word_processor = LineProcessor()

# Case 1: Term in quotes, is preceded by word|term|phrase or :,.^
# and has item from TRIGGER_LIST after itself.
# Fetch term along with quotes to be able to extract multiple terms,
# e.g.: the words "person" and "whoever" include
TRIGGER_WORDS_PTN = r"""
(?:(?:word|term|phrase)s?\s+|[:,\.]\s*|^)
['"“].{{1,{max_term_chars}}}['"”]\s*
(?:{trigger_list})[\s,]""".format(
    max_term_chars=MAX_TERM_CHARS,
    trigger_list=join_collection(ALL_TRIGGER_LIST))
TRIGGER_WORDS_PTN_RE = re.compile(TRIGGER_WORDS_PTN, re.IGNORECASE | re.UNICODE | re.DOTALL | re.MULTILINE | re.VERBOSE)

EXTRACT_PTN = r"""
“(.+?)“|
"(.+?)"|
'(.+?)'
"""
EXTRACT_PTN_RE = re.compile(EXTRACT_PTN, re.UNICODE | re.DOTALL | re.MULTILINE | re.VERBOSE)

ARTICLES = ['the', 'a', 'an']

# Case 2. Term inside quotes and brackets (the "Term") or ("Term")
PAREN_QUOTE_PTN = r"""\((?:each(?:,)?\s+)?(?:(?:{articles})\s+)?['"“](.{{1,{max_term_chars}}}?)\.?['"”]\)""" \
    .format(articles=join_collection(ARTICLES), max_term_chars=MAX_TERM_CHARS)
PAREN_QUOTE_PTN_RE_OPTIONS = re.IGNORECASE | re.UNICODE | re.DOTALL | re.MULTILINE | re.VERBOSE

# Case 2.5. Term inside brackets (TERM) or (Term), starts with uppercase
PAREN_PTN = r"""\((?:E|each(?:,)?\s+)?(?:(?:{articles})\s+)?([A-Z][^\)]{{1,{max_term_chars}}}?)\.?\)""" \
    .format(articles=join_collection(ARTICLES), max_term_chars=MAX_TERM_CHARS)
PAREN_PTN_RE_OPTIONS = re.UNICODE | re.DOTALL | re.MULTILINE | re.VERBOSE

# Case 3. Term is without quotes, is preceded by word|term|phrase or :,.^
# and has TRIGGER_LIST item after itself.
# e.g.: "Revolving Loan Commitment means…"; "LIBOR Rate shall mean…"
# false positive: "This Borrower Joiner Agreement to the extent signed signed and delivered by means of a facsimile..."
NOUN_PTN_BASE = r"""
(
    (?:[A-Z][-A-Za-z']*(?:\s*[A-Z][-A-Za-z']*){{0,{max_term_tokens}}})
    |
    (?:[A-Z][-A-Za-z'])
)
""".format(max_term_tokens=MAX_TERM_TOKENS)

NOUN_PTN = r"""
(?:^|\s)
(?:
    {noun_ptn_base}
    |
    "{noun_ptn_base}"
    |
    “{noun_ptn_base}”
)
\s+(?=(?:{trigger_list})\W)
""".format(noun_ptn_base=NOUN_PTN_BASE,
           trigger_list="|".join([w.replace(" ", r"\s+") for w in STRONG_TRIGGER_LIST]))

NOUN_PTN_RE = re.compile(NOUN_PTN, re.UNICODE | re.DOTALL | re.MULTILINE | re.VERBOSE)

NOUN_ANTI_PTN = r"""the\s*"""
NOUN_ANTI_PTN_RE = re.compile(NOUN_ANTI_PTN, re.IGNORECASE | re.UNICODE | re.DOTALL | re.MULTILINE | re.VERBOSE)

# Case 4: Term inside quotes is preceded by word|term|phrase or :,.^
# and has a colon after itself.
# e.g.: "'Term': definition"
COLON_PTN = r"""((['](.{{1,{max_term_chars}}})['])|""" \
            r"""(["](.{{1,{max_term_chars}}})["])|""" \
            r"""([“](.{{1,{max_term_chars}}})[”]))""" \
            r""":[\s]""".format(max_term_chars=MAX_TERM_CHARS)

COLON_PTN_RE_OPTIONS = re.UNICODE | re.DOTALL | re.MULTILINE

# Case 5: phrase called|herein|herein as... + term in quotes
# e.g.: all of which are herein collectively called the "Insurances", any such bank being an "Approved Bank"
ANCHOR = ['called', 'herein', 'herein as', 'collectively(?:,)?', 'collectively as', 'individually(?:,)?',
          'individually as', 'together(?:,)?', 'together with', 'referred to as', 'being', 'shall be', 'definition as',
          'known as', 'designated as', 'hereinafter', 'hereinafter as', 'hereafter', 'hereafter as', 'its', 'our',
          'your', 'any of the foregoing,', 'in such capacity,', 'in this section,', 'in this paragraph,',
          r'in this \(noun\),', 'each such', 'this']
ANCHOR_QUOTES_PTN = r"""(?:(?:{anchor})\s+)(?:(?:{articles})\s+)?['"“](.{{1,{max_term_chars}}}?)['"”]""" \
    .format(anchor=join_collection(ANCHOR), articles=join_collection(ARTICLES), max_term_chars=MAX_TERM_CHARS)
ANCHOR_QUOTE_RE_OPTIONS = re.IGNORECASE | re.UNICODE | re.DOTALL | re.MULTILINE | re.VERBOSE

# Case 6: phrase such|any such|together + subject + term in quotes
# e.g.: such earlier date, the "End Date", any such event, an "Event of Default"
ANCHOR = ['such', 'any such', 'together']
ANCHOR_SUBJECT_QUOTES_PTN = r"(?:(?:{anchor})\s+?)(?:.{{1,{max_term_chars}}}\s+?)(?:(?:{articles})\s+)?" \
                            r"(('(.{{1,{max_term_chars}}}?)')|" \
                            r"(\"(.{{1,{max_term_chars}}}?)\")|" \
                            r"(“(.{{1,{max_term_chars}}}?)”))" \
    .format(anchor=join_collection(ANCHOR), articles=join_collection(ARTICLES), max_term_chars=MAX_TERM_CHARS)
ANCHOR_SUBJECT_QUOTES_RE_OPTIONS = re.IGNORECASE | re.UNICODE | re.DOTALL | re.MULTILINE | re.VERBOSE

TRIGGER_QUOTED_DEFINITION_PATTERN = r"""['"“][^'"“]{{1,{max_term_chars}}}['"”]""".format(max_term_chars=MAX_TERM_CHARS)
TRIGGER_QUOTED_DEFINITION_RE = re.compile(TRIGGER_QUOTED_DEFINITION_PATTERN, re.DOTALL)

QUOTED_DEFINITION_RE_PARAMS = [
    (PAREN_PTN, PAREN_PTN_RE_OPTIONS),
    (PAREN_QUOTE_PTN, PAREN_QUOTE_PTN_RE_OPTIONS),
    (COLON_PTN, COLON_PTN_RE_OPTIONS),
    (ANCHOR_QUOTES_PTN, ANCHOR_QUOTE_RE_OPTIONS),
    (ANCHOR_SUBJECT_QUOTES_PTN, ANCHOR_SUBJECT_QUOTES_RE_OPTIONS)
]

QUOTED_DEFINITION_RE = [re.compile(template, options) for template, options in QUOTED_DEFINITION_RE_PARAMS]

QUOTED_TEXT_RE = re.compile("([\"'“„])(?:(?=(\\\\?))\\2.)+?\\1", re.UNICODE | re.IGNORECASE | re.DOTALL)

# used for replacing multiple spaces by single ones
SPACES_RE = re.compile(r'\s+')

# non significant parts of speech
# if defined term consists on NON_SIG_POSes only - this is not
# a definition (like "as is", "this" etc)
NON_SIG_POS = {'CC', 'CD', 'DT', 'EX', 'IN', 'LS',
               'MD', 'PDT', 'POS', 'PRP$', 'RB', 'RBR',
               'RBS', 'RP', 'TO', 'VBZ', 'WDT', 'WP', 'WP$', 'WRB',
               '.', ',', ':', '-', ';', ')', '(', ']', '{', '}'
               '[', '*', '/', '\\', '"', '\'', '!', '?', '%',
               '$', '^', '&', '@'}
# PDT predeterminer ‘all the kids’
# POS possessive ending parent’s
# RB adverb very, silently,
# RBR adverb, comparative better
# RBS adverb, superlative best

# punctuation POS that we have to skip while, e.g., removing introductory words
PUNCTUATION_POS = {'``', '\t'}.union(SpecialCharacters.punctuation)

# used for stripping strings
PUNCTUATION_STRIP_STR = ''.join(PUNCTUATION_POS)

# this punctuation should be removed from definition's name
STRIP_PUNCT_SYMBOLS = ',:-'


ABBREVIATION_PTRN = '|'.join([a.replace('.', '\\.') for a
                              in EnLanguageTokens.abbreviations])
# if the term ends with abbreviations, last dot won't be trimmed
ABBREVIATION_ENDING_RE = re.compile(f'({ABBREVIATION_PTRN})$')

# split one phrase containing several definitions into definitions
SPLIT_SUBDEFINITIONS_PTRN = r'''["“](?:[^"“]{{1,{max_term_chars}}})["“]'''.format(
    max_term_chars=MAX_TERM_CHARS)

SPLIT_SUBDEFINITIONS_RE = re.compile(SPLIT_SUBDEFINITIONS_PTRN, re.DOTALL)


def get_definition_list_in_sentence(sentence_coords: Tuple[int, int, str],
                                    decode_unicode=True) -> List[DefinitionCaught]:
    """
        Find possible definitions in natural language in a single sentence.
        :param sentence_coords: sentence, sentence start, end
        :param decode_unicode:
        :return:
        """
    definitions = []  # type: List[DefinitionCaught]
    sentence = sentence_coords[2]
    # unify quotes and braces
    # replace excess braces with ' ' so the str length will remain the same
    sentence = TextBeautifier.unify_quotes_braces(sentence,
                                                  empty_replacement=' ')
    sent_start = sentence_coords[0]
    result = set()  # type: Set[Tuple[str, int, int]]

    # it really transforms string, e.g. replaces “ with "
    if decode_unicode:
        sentence = unidecode.unidecode(sentence)
        sentence_coords = sentence_coords[0], sentence_coords[1], sentence

    # case 1
    for item in TRIGGER_WORDS_PTN_RE.finditer(sentence):
        result.update(regex_matches_to_word_coords(EXTRACT_PTN_RE, item.group(), item.start() + sent_start))

    # case 3
    mts = regex_matches_to_word_coords(NOUN_PTN_RE, sentence, sent_start)
    mts = [i for i in mts if not NOUN_ANTI_PTN_RE.fullmatch(i[0])]
    mts = [m for m in mts if m[0].lower().strip(' ,;.') not in EnLanguageTokens.pronouns]
    if len(mts) > 0:
        result.update(mts)

    # cases 2, 4, 5, 6
    for _ in TRIGGER_QUOTED_DEFINITION_RE.finditer(sentence):
        for quoted_definition_re in QUOTED_DEFINITION_RE:
            result.update(regex_matches_to_word_coords(quoted_definition_re, sentence, sent_start))
        break

    # make definitions out of entries
    for term, start, end in result:
        term_cleared = TextBeautifier.strip_pair_symbols((term, start, end))
        term_cleared = trim_defined_term(term_cleared[0], term_cleared[1], term_cleared[2])
        was_quoted = term_cleared[3]

        if PICK_DEFINITION_FROM_QUOTES:
            term, start, end = term_cleared[0], term_cleared[1], term_cleared[2]

        if not term_cleared[0]:
            continue

        term, start, end = TextBeautifier.unify_quotes_braces_coords(
            term, start, end)

        # check the term is not empty
        if len(term.strip(PUNCTUATION_STRIP_STR)) == 0:
            continue

        # returns [('word', 'token', (word_start, word_end)), ...] ...
        term_pos = list(SpanTokenizer.get_token_spans(term))
        if does_term_are_service_words(term_pos):
            continue

        term_wo_intro = IntroductoryWordsDetector.remove_term_introduction(
            term, term_pos)
        if term_wo_intro != term:
            term = TextBeautifier.strip_pair_symbols(term_wo_intro)
        if not term:
            continue

        # check the term is not too long
        max_words_per_definition = MAX_TERM_TOKENS
        if was_quoted:
            max_words_per_definition = MAX_QUOTED_TERM_TOKENS

        words_in_term = sum(1 for w in word_processor.split_text_on_words(term_cleared[0])
                            if not w.is_separator)
        quotes_in_text = get_quotes_count_in_string(term_cleared[0])
        possible_definitions = quotes_in_text // 2 if quotes_in_text > 1 else 1
        possible_tokens_count = max_words_per_definition * possible_definitions
        if words_in_term > possible_tokens_count:
            continue

        split_definitions_lst = split_definitions_inside_term(
            term, sentence_coords, start, end)

        for definition, s, e in split_definitions_lst:
            definition, s, e = TextBeautifier.strip_pair_symbols((definition, s, e))
            definitions.append(DefinitionCaught(definition, sentence, (s, e,)))

    return definitions


def split_definitions_inside_term(term: str,
                                  src_with_coords: Tuple[int, int, str],
                                  term_start: int,
                                  term_end: int) -> List[Tuple[str, int, int]]:
    """
    The whole phrase can be considered definition ("MSRB", "we", "us" or "our"),
    but in fact the phrase can be a collection of definitions.
    Here we split definition phrase to a list of definitions.

    Source string could be pre-processed, that's why we search for each
    sub-phrase's coordinates (PhrasePositionFinder)
    :param term: a definition or, probably, a set of definitions ("MSRB", "we", "us" or "our")
    :param src_with_coords: a sentence (probably), containing the term + its coords
    :param term_start: "term" start coordinate within the source sentence
    :param term_end: "term" end coordinate within the source sentence
    :return: [(definition, def_start, def_end), ...]
    """
    src_start = src_with_coords[0]
    src_text = src_with_coords[2]

    matches = [m.group() for m in SPLIT_SUBDEFINITIONS_RE.finditer(term)]
    if len(matches) < 2:
        matches = [term]

    match_coords = PhrasePositionFinder.find_phrase_in_source_text(
        src_text, matches, term_start - src_start, term_end - src_start)

    if len(match_coords) < len(matches):
        return [(term, term_start, term_end,)]

    match_coords = [(m[0], m[1] + src_start, m[2] + src_start, ) for m in match_coords]

    return match_coords


def does_term_are_service_words(term_pos: List[Tuple[str, str, int, int]]) -> bool:
    """
    Does term consist of service words only?
    """
    for _, pos, _, _ in term_pos:
        if pos not in NON_SIG_POS:
            return False
    return True


def trim_defined_term(term: str, start: int, end: int) -> \
        Tuple[str, int, int, bool]:
    """
    Remove pair of quotes / brackets framing text
    Replace N-grams of spaces with single spaces
    Replace line breaks with spaces
    :param term: a phrase that may contain excess framing symbols
    :param start: original term's start position, may be changed
    :param end: original term's end position, may be changed
    :return: updated term, start, end and the flag indicating that the whole phrase was inside quotes
    """
    was_quoted = False

    # pick text from quotes
    # pick text from quotes
    quoted_parts = [m.group() for m in QUOTED_TEXT_RE.finditer(term)]
    if len(quoted_parts) == 1:
        term = quoted_parts[0].strip('''\"'“„''')
        was_quoted = True

    orig_term_len = len(term)
    orig_term_quotes = count_sequence_matches(term, lambda c: c in TextBeautifier.QUOTES)
    term, start, end = TextBeautifier.strip_pair_symbols((term, start, end))
    if len(term) < orig_term_len:
        # probably we removed quotes
        updated_term_quotes = count_sequence_matches(term, lambda c: c in TextBeautifier.QUOTES)
        was_quoted = was_quoted or orig_term_quotes - updated_term_quotes > 1

    term = term.replace('\n', ' ')
    term = SPACES_RE.sub(' ', term)

    term, start, end = TextBeautifier.strip_string_coords(
        term, start, end, STRIP_PUNCT_SYMBOLS)

    # strip all dots or just left one (if ends with abbreviation)
    ends_with_abbr = ABBREVIATION_ENDING_RE.search(term)
    if not ends_with_abbr:
        term, start, end = TextBeautifier.strip_string_coords(
            term, start, end, '.')
    else:
        term, start, end = TextBeautifier.lstrip_string_coords(
            term, start, end, '.')



    return term, start, end, was_quoted


def filter_definitions_for_self_repeating(definitions: List[DefinitionCaught]) -> List[DefinitionCaught]:
    """
    :param definitions:
    :return: excludes definitions that are "overlapped", leaves unique definitions only
    """
    for i in range(0, len(definitions)):
        a = definitions[i]
        if not a.name:
            continue
        for j in range(i + 1, len(definitions)):
            b = definitions[j]
            consumes = a.does_consume_target(b)
            if consumes == 1:
                b.name = None
            elif consumes == -1:
                a.name = None

    return [d for d in definitions if d.name is not None]


def get_quotes_count_in_string(text: str) -> int:
    """
    :param text: text to calculate quotes within
    :return: calculates count of quotes within the text passed
    """
    c = Counter(text)
    return sum(filter(None, [c['"'], c['”']]))


def regex_matches_to_word_coords(pattern: Pattern[str],
                                 text: str, phrase_start: int = 0) -> List[Tuple[str, int, int]]:
    """
    :param pattern: pattern for searching for matches within the text
    :param text: text to search for matches
    :param phrase_start: a value to be add to start / end
    :return: tuples of (match_text, start, end) out of the regex (pattern) matches in text
    """
    return [(m.group(), m.start() + phrase_start, m.end() + phrase_start)
            for m in pattern.finditer(text)]



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# To use a consistent encoding
from os import path

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
try:
    with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
        long_description = f.read()
except Exception as e:
    long_description = "LexPredict LexNLP: A swiss-army knife library built for working with real, unstructured legal text."

setup(
    name='lexnlp',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='1.4.0',

    description='LexPredict LexNLP',
    long_description=long_description,

    # The project's main homepage.
    url='https://contraxsuite.com',

    # Author details
    author='ContraxSuite, LLC',
    author_email='support@contraxsuite.com',

    # Choose your license
    license='AGPL',

    # version ranges for supported Python distributions
    python_requires='~=3.6',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Information Technology',
        'Intended Audience :: Legal Industry',
        'Intended Audience :: Developers',

        # Pick your license as you wish (should match "license" above)
        # 'License :: OSI Approved :: MIT License',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'License :: Other/Proprietary License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.6',

        # Topics
        'Natural Language :: English',
        'Topic :: Office/Business',
        'Topic :: Text Processing :: Linguistic',

    ],

    # What does your project relate to?
    keywords='legal contract document analytics nlp ml machine learning natural language',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['*tests*']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    # py_modules=['lexnlp'],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        # 'datefinder @ https://github.com/LexPredict/datefinder/archive/master.zip',
        'datefinder-lexpredict==0.6.2',
        'dateparser==0.7.0',
        'gensim==3.4.0',
        'nltk==3.4.5',
        'num2words==0.5.7',
        'pandas==0.23.4',
        'pycountry==18.5.26',
        'regex==2017.9.23',
        'reporters-db==1.0.12.1',
        'requests==2.22.0',
        'scipy==1.0.0',
        'scikit-learn==0.21.3',
        'typing==3.6.2',
        'Unidecode==0.4.21',
        'us==1.0.0'
    ],
    dependency_links=[
        'git+https://github.com/LexPredict/datefinder.git@master#egg=datefinder-0.6.2'
    ],

    # Install any data files from packages.
    # The data files must be specified via the distutils’ MANIFEST.in file.
    include_package_data=True,

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': ['pytest>=2.8.5', 'mock', 'pytz>=2015.7'],
        'test': ['pytest>=2.8.5', 'mock', 'pytz>=2015.7'],
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    # package_data={
    #     'sample': ['package_data.dat'],
    # },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    # entry_points={
    #     'console_scripts': [
    #         'sample=sample:main',
    #     ],
    # },
)



"""Entity extraction for English using Stanford Named Entity Recognition (NER).

This module implements basic entity extraction functionality in English relying on the
pre-trained Stanford NLP NER classifiers.

Todo:
  * Better define interface for sentences vs. raw text
  * Standardize generator vs list
"""

# Change the path according to your system
import os
import string

from typing import Generator

from nltk.tag import StanfordNERTagger

from lexnlp.config.stanford import STANFORD_NER_PATH
from lexnlp.extract.en.utils import strip_unicode_punctuation
from lexnlp.nlp.en.segments.sentences import get_sentence_list
from lexnlp.nlp.en.stanford import get_tokens_list

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


# Setup Stanford NER configuration
STANFORD_NER_FILE = os.path.join(STANFORD_NER_PATH, "stanford-ner.jar")
STANFORD_NER_MODEL_MAP = {"english": "english.all.3class.distsim.crf.ser.gz",
                          "english7": "english.muc.7class.distsim.crf.ser.gz"}


def get_model_file(language):
    """
    Return the appropriate model file for each language.
    :param language:
    :return:
    """
    return os.path.join(STANFORD_NER_PATH, "classifiers", STANFORD_NER_MODEL_MAP[language])


try:
    STANFORD_NER_TAGGER = StanfordNERTagger(get_model_file("english"), STANFORD_NER_FILE, encoding='utf-8')
except LookupError:
    STANFORD_NER_TAGGER = None


def get_persons(text, strict=False, return_source=False, window=2) -> Generator:
    """
    Get persons from text using Stanford libraries.
    :param window:
    :param return_source:
    :param strict:
    :param text:
    :return:
    """
    # Iterate through sentences
    for sentence in get_sentence_list(text):
        # Tag sentence
        sentence_pos = STANFORD_NER_TAGGER.tag(get_tokens_list(text))

        # Iterate through chunks
        names = []
        last_person_pos = None
        for i, token in enumerate(sentence_pos):
            # Check label
            if token[1] == 'PERSON':
                if not strict and last_person_pos is not None and (i - last_person_pos) < window:
                    names[-1] += " " + token[0]
                else:
                    names.append(token[0])
                last_person_pos = i
            else:
                if token[0] in [".", ","]:
                    if not strict and last_person_pos is not None and (i - last_person_pos) < window:
                        names[-1] += (" " if token[0] not in string.punctuation else "") + token[0]
                        last_person_pos = i

        # Cleanup and yield
        for name in names:
            name = strip_unicode_punctuation(name).strip(string.punctuation).strip(string.whitespace)
            if return_source:
                yield name, sentence
            else:
                yield name


def get_organizations(text, strict=False, return_source=False, window=2) -> Generator:
    """
    Get organizations from text using Stanford libraries.
    :param window:
    :param return_source:
    :param strict:
    :param text:
    :return:
    """
    # Iterate through sentences
    for sentence in get_sentence_list(text):
        # Tag sentence
        sentence_pos = STANFORD_NER_TAGGER.tag(get_tokens_list(text))

        # Iterate through chunks
        orgs = []
        last_org_pos = None
        for i, token in enumerate(sentence_pos):
            # Check label
            if token[1] == 'ORGANIZATION':
                if not strict and last_org_pos is not None and (i - last_org_pos) < window:
                    orgs[-1] += " " + token[0]
                else:
                    orgs.append(token[0])
                last_org_pos = i
            else:
                if token[0] in [".", ","]:
                    if not strict and last_org_pos is not None and (i - last_org_pos) < window:
                        orgs[-1] += (" " if token[0] not in string.punctuation else "") + token[0]
                        last_org_pos = i

        # Cleanup and yield
        for org in orgs:
            org = strip_unicode_punctuation(org).strip(string.punctuation).strip(string.whitespace)
            if return_source:
                yield org, sentence
            else:
                yield org


def get_locations(text, strict=False, return_source=False, window=2) -> Generator:
    """
    Get locations from text using Stanford libraries.
    :param window:
    :param return_source:
    :param strict:
    :param text:
    :return:
    """
    # Iterate through sentences
    for sentence in get_sentence_list(text):
        # Tag sentence
        sentence_pos = STANFORD_NER_TAGGER.tag(get_tokens_list(text))

        # Iterate through chunks
        locations = []
        last_loc_pos = None
        for i, token in enumerate(sentence_pos):
            # Check label
            if token[1] == 'LOCATION':
                if not strict and last_loc_pos is not None and (i - last_loc_pos) < window:
                    locations[-1] += (" " if not token[0].startswith("'") else "") + token[0]
                else:
                    locations.append(token[0])
                last_loc_pos = i
            else:
                if token[0] in [".", ","]:
                    if not strict and last_loc_pos is not None and (i - last_loc_pos) < window:
                        locations[-1] += (" " if token[0] not in string.punctuation and not token[0].startswith(
                            "'") else "") + token[0]
                        last_loc_pos = i

        # Cleanup and yield
        for location in locations:
            location = strip_unicode_punctuation(location).strip(string.punctuation).strip(string.whitespace)
            if return_source:
                yield location, sentence
            else:
                yield location



import os

from lexnlp.extract.common.language_dictionary_reader import LanguageDictionaryReader

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class EnLanguageTokens:
    abbreviations = {'nr.', 'abs.', 'no.', 'act.', 'inc.', 'p.', 'Inc.'}
    articles = ['a', 'the', 'an']
    conjunctions = ['for', 'and', 'nor', 'but', 'or', 'yet', 'so']
    pronouns = {'I', 'he', 'she', 'we', 'you', 'they'}

    @staticmethod
    def init():
        abr_file_path = os.path.join(os.path.dirname(__file__),
                                     'data/abbreviations.txt')
        if os.path.isfile(abr_file_path):
            file_set = LanguageDictionaryReader.read_str_set(abr_file_path)
            EnLanguageTokens.abbreviations = \
                EnLanguageTokens.abbreviations.union(file_set)

        prn_file_path = os.path.join(os.path.dirname(__file__),
                                     'data/pronouns.txt')
        if os.path.isfile(prn_file_path):
            file_set = LanguageDictionaryReader.read_str_set(prn_file_path)
            EnLanguageTokens.pronouns = \
                EnLanguageTokens.pronouns.union(file_set)


EnLanguageTokens.init()



"""Amount extraction for German.

This module implements amount extraction functionality in German.

Todo:
  * Improved unit tests and case coverage
"""

# pylint: disable=broad-except

import string
from typing import Generator

import nltk
import regex as re
from num2words import num2words, CONVERTER_CLASSES

from lexnlp.extract.common.annotations.amount_annotation import AmountAnnotation

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


CURRENCY_SYMBOL_MAP = {
    "$": "USD",
    "€": "EUR",
    "¥": "JPY",
    "£": "GBP",
    "₠": "EUR",
    "₨": "INR",
    "₹": "INR",
    "₺": "TRY",
    "元": "CNY",
    "₽": "RUB",
    # "¢": None,
    "₩": "KRW",
}
CURRENCY_PREFIX_MAP = {
    "chf": "CHF",
    "rmb": "CNY",
}
allowed_prev_units = list(CURRENCY_SYMBOL_MAP) + list(CURRENCY_PREFIX_MAP)

grammar = r"""
    NBAR:
        {<NN.*|JJ>*<NN.*>}  # Nouns and Adjectives, terminated with Nouns
    NP:
        {<NBAR>}
        {<NBAR><IN><NBAR>}  # Above, connected with in/of/etc...
"""
chunker = nltk.RegexpParser(grammar)


def get_np(text) -> Generator:
    tokens = nltk.word_tokenize(text)
    pos_tokens = nltk.tag.pos_tag(tokens)
    chunks = chunker.parse(pos_tokens)
    for subtree in chunks.subtrees(filter=lambda t: t.label() == 'NP'):
        np = ' '.join([i[0] for i in subtree.leaves()])
        yield np


class AmountParserDE(object):
    QUARTER = 'viertel'

    def __init__(self):

        self.language = 'de'

        N2W_CONFIG = CONVERTER_CLASSES[self.language]

        self.BIG_NUMBERS_EXPONENT = [3, 6, 9, 12]
        self.ONE = N2W_CONFIG.low_numwords[-2]
        self.HUNDRED = dict(N2W_CONFIG.mid_numwords)[100]

        UNIQUE_NUMBERS = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
                          20, 30, 40, 50, 60, 70, 80, 90,
                          100, 1000, 1000000, 1000000000, 1000000000000]
        UNIQUE_NUMBERS_MAP = {num2words(n, ordinal=True, lang=self.language).replace('eine ', ''): n
                              for n in UNIQUE_NUMBERS}
        UNIQUE_NUMBERS_MAP.update(
            {num2words(n, lang=self.language).replace('eine ', ''): n for n in UNIQUE_NUMBERS})
        UNIQUE_NUMBERS_MAP.update(
            {'ein': 1,
             'eine': 1,
             'einen': 1,
             'einhalb': 0.5,
             'millionen': 1000000,
             'millionenste': 1000000,
             'milliarden': 1000000000,
             'milliardenste': 1000000000})

        self.UNIQUE_NUMBERS_MAP = UNIQUE_NUMBERS_MAP

        self.MAGNITUDE_MAP = {num2words(10 ** n, lang=self.language).replace('eine ', ''): 10 ** n
                              for n in self.BIG_NUMBERS_EXPONENT}
        self.MAGNITUDE_MAP.update(
            {'millionen': 1000000,
             'millionenste': 1000000,
             'milliarden': 1000000000,
             'milliardenste': 1000000000,
             'halbe': 0.5,
             'k': 1000,
             'm': 1000000,
             'b': 1000000000})

        unique_number_list = list(self.UNIQUE_NUMBERS_MAP.keys())
        unique_number_list.sort(key=len, reverse=True)
        self.UNIQUE_NUMBER_SPLIT_RE = re.compile(r'({}|\s+)'.format('|'.join(unique_number_list)))

        self.NUM_PTN = r"""
        (?:
        (?:[\.\d][\d\.,]*\s+|\W|^)
        (?:(?:(?:in)?viertel|halbe|{written_unique_numbers}|und)\s?)+(?:\W|$)|
        (?:[\.\d][\d\.,\s]*)
        )
        """.format(written_unique_numbers='|'.join(unique_number_list))
        self.NUM_PTN_RE = re.compile(self.NUM_PTN,
                                     re.IGNORECASE | re.MULTILINE | re.DOTALL | re.VERBOSE)

        self.NON_WRIT_RE = re.compile(r'[\d\.,\s]+')

        self.MIXED_WRIT_RE = re.compile(r'(^[\d\.]*)(.+)', re.DOTALL)

        self.QUARTER_RE = re.compile(r'(?:\s*und\s+)?(ein|eine|zwei|drei)\s*viertel')

        self.WRONG_FULLMATCH_RE = re.compile(r'\W*und\W*|\W+', re.IGNORECASE | re.MULTILINE | re.DOTALL)

    def cleanup(self, text):
        text = text.lower().replace(',', '.').replace('-', ' ').strip(string.whitespace).rstrip(
            string.punctuation + string.whitespace)
        if not (text.startswith('.') and text[1].isdigit()):
            text = text.lstrip(string.punctuation + string.whitespace)
        return text

    def split(self, text):
        return [i for i in self.UNIQUE_NUMBER_SPLIT_RE.split(text) if i not in ['', ' ']]

    def text2num(self, s):
        """
        Convert written amount into integer/float.
        :param s: written number
        :param search_fraction: extract fraction
        :return: integer/float
        """
        n = 0
        g = 0
        s = self.cleanup(s)

        # if only number or float in string
        if self.NON_WRIT_RE.fullmatch(s):
            return float(s.replace(' ', ''))

        # if written number has integer/float prefix: "25 million", "2.035 thousand tons"
        if self.MIXED_WRIT_RE.search(s):
            p, s = self.MIXED_WRIT_RE.search(s).groups()
            g = float(p.rstrip(string.punctuation + string.whitespace)) if p else 0

        d = 0
        # TODO: extract fractions, half, quarter

        # convert quarters
        if self.QUARTER_RE and self.QUARTER_RE.search(s):
            nu = self.QUARTER_RE.search(s).groups()[0]
            d = self.text2num(nu) / 4
            s = self.QUARTER_RE.sub('', s)

        # process
        a = self.split(s)

        for w in a:
            if w == 'und':
                continue

            x = self.UNIQUE_NUMBERS_MAP.get(w, None)
            if self.HUNDRED in w and g != 0:
                g *= 100

            elif w in self.MAGNITUDE_MAP:
                x = self.MAGNITUDE_MAP.get(w, None)
                if w == 'halbe':
                    g = 0.5
                    continue
                n += (g or 1) * x
                g = 0

            elif x is None:
                raise RuntimeError('Unknown number: ' + w)

            else:
                g += x

        return n + g + d

    def parse(self, text: str, return_sources=False,
              extended_sources=True, float_digits=4) -> Generator:
        """
        Find possible amount references in the text.
        :param text: text
        :param return_sources: return amount AND source text
        :param extended_sources: return data around amount itself
        :param float_digits: round float to N digits, don't round if None
        :return: list of amounts
        """
        for ant in self.parse_annotations(text, return_sources, float_digits):
            if not return_sources:
                yield ant.value
            else:
                if extended_sources:
                    yield (ant.value, ant.text, ant.coords)
                else:
                    yield (ant.value, ant.text)

    def parse_annotations(self,
                          text: str,
                          float_digits=4,
                          return_sources=True) \
            -> Generator[AmountAnnotation, None, None]:
        """
        Find possible amount references in the text.
        :param text: text
        :param return_sources: return amount AND source text
        :param extended_sources: return data around amount itself
        :param float_digits: round float to N digits, don't round if None
        :return: list of amounts
        """
        for match in self.NUM_PTN_RE.finditer(text):
            found_item = match.group()
            if self.WRONG_FULLMATCH_RE.fullmatch(found_item):
                continue
            try:
                amount = self.text2num(found_item)
            except Exception as e:
                print(e)
                continue
            if amount is None:
                continue
            if isinstance(amount, float) and float_digits:
                amount = round(amount, float_digits)

            ant = AmountAnnotation(coords=match.span(),
                                   value=amount,
                                   locale=self.language)

            if return_sources:
                unit = ''
                next_text = text[match.span()[1]:]
                if next_text:
                    for np in get_np(next_text):
                        if next_text.startswith(np):
                            unit = np
                    if unit:
                        found_item = ' '.join([found_item.strip(), unit])
                if not unit:
                    prev_text = text[:match.span()[0]]
                    prev_text_tags = nltk.word_tokenize(prev_text)
                    if prev_text_tags and prev_text_tags[-1].lower() in allowed_prev_units:
                        sep = ' ' if text[match.span()[0] - 1] == ' ' else ''
                        found_item = sep.join([prev_text_tags[-1], found_item.rstrip()])

                ant.text = found_item.strip()
            yield ant


get_amounts = AmountParserDE().parse


get_amount_annotations = AmountParserDE().parse_annotations


def get_amount_list(*args, **kwargs):
    return list(get_amounts(*args, **kwargs))
